# Copyright (c) 2018-2020, Manfred Moitzi
# License: MIT License
from typing import TYPE_CHECKING, List, Union, Optional, Tuple
import copy
import logging
from collections import namedtuple

from ezdxf.lldxf import const
from ezdxf.lldxf.attributes import (
    DXFAttr, DXFAttributes, DefSubclass, XType, group_code_mapping,
)
from ezdxf.lldxf.tags import Tags
from ezdxf.math import Vec3, NULLVEC, X_AXIS, Y_AXIS, Z_AXIS, Matrix44
from ezdxf import colors
from .dxfentity import base_class, SubclassProcessor
from .dxfobj import DXFObject
from .dxfgfx import DXFGraphic, acdb_entity

from .factory import register_entity
from .objectcollection import ObjectCollection

if TYPE_CHECKING:
    from ezdxf.eztypes import TagWriter, Drawing, DXFNamespace, DXFTag

__all__ = ['MultiLeader', 'MLeaderStyle', 'MLeaderStyleCollection']

logger = logging.getLogger('ezdxf')

# DXF Examples:
# "D:\source\dxftest\CADKitSamples\house design for two family with common staircasedwg.dxf"
# "D:\source\dxftest\CADKitSamples\house design.dxf"

# How to render MLEADER: https://atlight.github.io/formats/dxf-leader.html
# DXF reference:
# http://help.autodesk.com/view/OARX/2018/ENU/?guid=GUID-72D20B8C-0F5E-4993-BEB7-0FCF94F32BE0

acdb_mleader = DefSubclass('AcDbMLeader', {
    'version': DXFAttr(270, default=2),
    'style_handle': DXFAttr(340),

    # Theory: Take properties from MLEADERSTYLE,
    # except explicit overridden here:
    'property_override_flags': DXFAttr(90),
    # Bit coded flags:
    # 1 << 0 = leader_type
    # 1 << 1 = leader_line_color
    # 1 << 2 = leader_linetype_handle
    # 1 << 3 = leader_lineweight
    # 1 << 4 = landing_flag
    # 1 << 5 = landing_gap ???
    # 1 << 6 = dogleg_flag
    # 1 << 7 = dogleg_length
    # 1 << 8 = arrow_head_handle
    # 1 << 9 = arrow_head_size
    # 1 << 10 = content_type
    # 1 << 11 = text_style_handle
    # 1 << 12 = text_left_attachment_type (of MTEXT)
    # 1 << 13 = text_angle_type (of MTEXT)
    # 1 << 14 = text_alignment_type (of MTEXT)
    # 1 << 15 = text_color (of MTEXT)
    # 1 << 16 = ??? Text height (of MTEXT) ???
    # 1 << 17 = text_frame_flag
    # 1 << 18 = ??? Enable use of default MTEXT (from MLEADERSTYLE)
    # 1 << 19 = block_record_handle
    # 1 << 20 = block_color
    # 1 << 21 = block_scale_vector
    # 1 << 22 = block_rotation
    # 1 << 23 = block_connection_type
    # 1 << 24 = ??? Scale ???
    # 1 << 25 = text_right_attachment_type (of MTEXT)
    # 1 << 26 = ??? Text switch alignment type (of MTEXT) ???
    # 1 << 27 = text_attachment_direction (of MTEXT)
    # 1 << 28 = text_top_attachment_type (of MTEXT)
    # 1 << 29 = Text_bottom_attachment_type (of MTEXT)

    'leader_type': DXFAttr(170, default=1),
    'leader_line_color': DXFAttr(91, default=colors.BY_BLOCK_RAW_VALUE),
    'leader_linetype_handle': DXFAttr(341),
    'leader_lineweight': DXFAttr(171, default=const.LINEWEIGHT_BYBLOCK),
    'has_landing': DXFAttr(290, default=1),
    'has_dogleg': DXFAttr(291, default=1),
    'dogleg_length': DXFAttr(41, default=8),  # depend on $MEASUREMENT?

    # no handle is default arrow 'closed filled':
    'arrow_head_handle': DXFAttr(342),
    'arrow_head_size': DXFAttr(42, default=4),  # depend on $MEASUREMENT?

    'content_type': DXFAttr(172, default=2),
    # 0 = None
    # 1 = Block content
    # 2 = MTEXT content
    # 3 = TOLERANCE content

    # Text Content:
    'text_style_handle': DXFAttr(343),
    'text_left_attachment_type': DXFAttr(173, default=1),
    # Values 0-8 are used for the left/right attachment
    # point (attachment direction is horizontal), values 9-10 are used for the
    # top/bottom attachment points (attachment direction is vertical).
    # Attachment point is:
    # 0 = top of top text line,
    # 1 = middle of top text line,
    # 2 = middle of text,
    # 3 = middle of bottom text line,
    # 4 = bottom of bottom text line,
    # 5 = bottom text line,
    # 6 = bottom of top text line. Underline bottom line
    # 7 = bottom of top text line. Underline top line,
    # 8 = bottom of top text line. Underline all content,
    # 9 = center of text (y-coordinate only),
    # 10 = center of text (y-coordinate only), and overline top/underline
    # bottom content.
    'text_right_attachment_type': DXFAttr(95),  # like 173

    'text_angle_type': DXFAttr(174, default=1),
    # 0 = text angle is equal to last leader line segment angle
    # 1 = text is horizontal
    # 2 = text angle is equal to last leader line segment angle, but potentially
    #     rotated by 180 degrees so the right side is up for readability.

    'text_alignment_type': DXFAttr(175, default=2),
    'text_color': DXFAttr(92, default=colors.BY_BLOCK_RAW_VALUE),
    'has_frame_text': DXFAttr(292, default=0),

    # Block Content:
    'block_record_handle': DXFAttr(344),
    'block_color': DXFAttr(93, default=colors.BY_BLOCK_RAW_VALUE),  # raw color
    'block_scale_vector': DXFAttr(10, xtype=XType.point3d,
                                  default=Vec3(1, 1, 1)),
    'block_rotation': DXFAttr(43, default=0),  # in radians!!!
    'block_connection_type': DXFAttr(176, default=0),
    # 0 = center extents
    # 1 = insertion point

    'is_annotative': DXFAttr(293, default=0),

    # REPEAT "arrow_heads": DXF R2007+
    # arrow_head_index: 94, ???
    # arrow_head_handle: 345
    # END "arrow heads"

    # REPEAT "block attribs" (ATTDEF): DXF R2007+
    # attrib_handle: 330
    # attrib_index: 177, sequential index of the label in the collection
    # attrib_width: 44
    # attrib_text: 302, collision with group code (302, "LEADER{") in context data
    # END "block attribs"

    # Text Content:
    'is_text_direction_negative': DXFAttr(294, default=0,
                                          dxfversion=const.DXF2007),
    'text_IPE_align': DXFAttr(178, default=0, dxfversion=const.DXF2007),
    'text_attachment_point': DXFAttr(179, default=1, dxfversion=const.DXF2007),
    # 1 = left
    # 2 = center
    # 3 = right

    'scale': DXFAttr(45, default=1, dxfversion=const.DXF2007),

    'text_attachment_direction': DXFAttr(
        271, default=0, dxfversion=const.DXF2010),
    # This defines whether the leaders attach to the left/right of the content
    # block/text, or attach to the top/bottom:
    # 0 = horizontal
    # 1 = vertical

    'text_bottom_attachment_direction': DXFAttr(
        272, default=9, dxfversion=const.DXF2010),
    # like 173, but
    # 9 = center
    # 10= underline and center

    'text_top_attachment_direction': DXFAttr(
        273, default=9, dxfversion=const.DXF2010),
    # like 173, but
    # 9 = center
    # 10= overline and center

    'leader_extend_to_text': DXFAttr(
        295, default=0, dxfversion=const.DXF2013),

})
acdb_mleader_group_codes = group_code_mapping(acdb_mleader)
CONTEXT_STR = 'CONTEXT_DATA{'
LEADER_STR = 'LEADER{'
LEADER_LINE_STR = 'LEADER_LINE{'
START_CONTEXT_DATA = 300
END_CONTEXT_DATA = 301
START_LEADER = 302
END_LEADER = 303
START_LEADER_LINE = 304
END_LEADER_LINE = 305


def compile_context_tags(data: List['DXFTag'],
                         stop_code: int) -> List[Union['DXFTag', List]]:
    def build_structure(tag: 'DXFTag',
                        stop: int) -> List[Union['DXFTag', List]]:
        collector = [tag]
        tag = next(tags)
        while tag.code != stop:
            if tag.code == START_LEADER:
                collector.append(build_structure(tag, END_LEADER))
            # Group code 304 is used also for MTEXT content, therefore always
            # test for group code AND and value string:
            elif tag.code == START_LEADER_LINE and tag.value == LEADER_LINE_STR:
                collector.append(build_structure(tag, END_LEADER_LINE))
            else:
                collector.append(tag)
            tag = next(tags)
        return collector

    tags = iter(data)
    return build_structure(next(tags), stop_code)


ArrowHeadData = namedtuple('ArrowHeadData', 'index, handle')
AttribData = namedtuple('AttribData', 'handle, index, width, text')


@register_entity
class MultiLeader(DXFGraphic):
    DXFTYPE = 'MULTILEADER'
    DXFATTRIBS = DXFAttributes(base_class, acdb_entity, acdb_mleader)
    MIN_DXF_VERSION_FOR_EXPORT = const.DXF2000

    def __init__(self):
        super().__init__()
        self.context = MultiLeaderContext()
        self.arrow_heads: List[ArrowHeadData] = []
        self.block_attribs: List[AttribData] = []

    def _copy_data(self, entity: 'MultiLeader') -> None:
        """ Copy leaders """
        entity.context = copy.deepcopy(self.context)
        entity.arrow_heads = copy.deepcopy(self.arrow_heads)
        entity.block_attribs = copy.deepcopy(self.block_attribs)

    def load_dxf_attribs(
            self, processor: SubclassProcessor = None) -> 'DXFNamespace':
        dxf = super().load_dxf_attribs(processor)
        if processor is None:
            return dxf
        tags = processor.subclass_by_index(2)
        context = self.extract_context_data(tags)
        if context:
            try:
                self.context = self.load_context(context)
            except const.DXFStructureError:
                logger.info(
                    f'Context structure error in entity MULTILEADER(#{dxf.handle})')

        self.arrow_heads = self.extract_arrow_heads(tags)
        self.block_attribs = self.extract_block_attribs(tags)

        processor.fast_load_dxfattribs(
            dxf, acdb_mleader_group_codes, subclass=tags, recover=True)
        return dxf

    @staticmethod
    def extract_context_data(tags: Tags) -> List['DXFTag']:
        start, end = None, None
        context_data = []
        for index, tag in enumerate(tags):
            if tag.code == START_CONTEXT_DATA:
                start = index
            elif tag.code == END_CONTEXT_DATA:
                end = index + 1

        if start and end:
            context_data = tags[start:end]
            # Remove context data!
            del tags[start: end]
        return context_data

    @staticmethod
    def load_context(data: List['DXFTag']) -> 'MultiLeaderContext':
        try:
            context = compile_context_tags(data, END_CONTEXT_DATA)
        except StopIteration:
            raise const.DXFStructureError
        else:
            return MultiLeaderContext.load(context)

    @staticmethod
    def extract_arrow_heads(data: Tags) -> List[ArrowHeadData]:
        def store_head():
            heads.append(ArrowHeadData(
                collector.get(94, 0),  # arrow head index
                collector.get(345, '0'),  # arrow head handle
            ))
            collector.clear()

        heads = []
        try:
            start = data.tag_index(94)
        except const.DXFValueError:
            return heads

        end = start
        collector = dict()
        for code, value in data.collect_consecutive_tags({94, 345}, start):
            end += 1
            collector[code] = value
            if code == 345:
                store_head()

        # Remove processed tags:
        del data[start: end]
        return heads

    @staticmethod
    def extract_block_attribs(data: Tags) -> List[AttribData]:
        def store_attrib():
            attribs.append(AttribData(
                collector.get(330, '0'),  # ATTDEF handle
                collector.get(177, 0),  # ATTDEF index
                collector.get(44, 1.0),  # ATTDEF width
                collector.get(302, ''),  # ATTDEF text (content)
            ))
            collector.clear()

        attribs = []
        try:
            start = data.tag_index(330)
        except const.DXFValueError:
            return attribs

        end = start
        collector = dict()
        for code, value in data.collect_consecutive_tags(
                {330, 177, 44, 302}, start):
            end += 1
            if code == 330 and len(collector):
                store_attrib()
            collector[code] = value
        if len(collector):
            store_attrib()

        # Remove processed tags:
        del data[start: end]
        return attribs

    def preprocess_export(self, tagwriter: 'TagWriter') -> bool:
        if self.context.is_valid:
            return True
        else:
            logger.debug(
                f'Ignore {str(self)} at DXF export, invalid context data.')
            return False

    def export_entity(self, tagwriter: 'TagWriter') -> None:
        def write_handle_if_exist(code: int, name: str):
            handle = dxf.get(name)
            if handle is not None:
                write_tag2(code, handle)

        super().export_entity(tagwriter)
        dxf = self.dxf
        version = tagwriter.dxfversion
        write_tag2 = tagwriter.write_tag2

        write_tag2(100, acdb_mleader.name)
        write_tag2(270, dxf.version)
        self.context.export_dxf(tagwriter)

        # Export common MLEADER tags:
        # Don't use dxf.export_dxf_attribs() - all attributes should be written
        # even if equal to the default value:
        write_tag2(340, dxf.style_handle)
        write_tag2(90, dxf.property_override_flags)
        write_tag2(170, dxf.leader_type)
        write_tag2(91, dxf.leader_line_color)
        write_tag2(341, dxf.leader_linetype_handle)
        write_tag2(171, dxf.leader_lineweight)
        write_tag2(290, dxf.has_landing)
        write_tag2(291, dxf.has_dogleg)
        write_tag2(41, dxf.dogleg_length)
        # arrow_head_handle is None for default arrow 'closed filled':
        write_handle_if_exist(342, 'arrow_head_handle')
        write_tag2(42, dxf.arrow_head_size)
        write_tag2(172, dxf.content_type)
        write_tag2(343, dxf.text_style_handle)  # mandatory!
        write_tag2(173, dxf.text_left_attachment_type)
        write_tag2(95, dxf.text_right_attachment_type)
        write_tag2(174, dxf.text_angle_type)
        write_tag2(175, dxf.text_alignment_type)
        write_tag2(92, dxf.text_color)
        write_tag2(292, dxf.has_frame_text)

        write_handle_if_exist(344, 'block_record_handle')
        write_tag2(93, dxf.block_color)
        tagwriter.write_vertex(10, dxf.block_scale_vector)
        write_tag2(43, dxf.block_rotation)
        write_tag2(176, dxf.block_connection_type)
        write_tag2(293, dxf.is_annotative)
        if version >= const.DXF2007:
            self.export_arrow_heads(tagwriter)
            self.export_block_attribs(tagwriter)
            write_tag2(294, dxf.is_text_direction_negative)
            write_tag2(178, dxf.text_IPE_align)
            write_tag2(179, dxf.text_attachment_point)
            write_tag2(45, dxf.scale)

        if version >= const.DXF2010:
            write_tag2(271, dxf.text_attachment_direction)
            write_tag2(272, dxf.text_bottom_attachment_direction)
            write_tag2(273, dxf.text_top_attachment_direction)

        if version >= const.DXF2013:
            write_tag2(295, dxf.leader_extend_to_text)

    def export_arrow_heads(self, tagwriter: 'TagWriter') -> None:
        for index, handle in self.arrow_heads:
            tagwriter.write_tag2(94, index)
            tagwriter.write_tag2(345, handle)

    def export_block_attribs(self, tagwriter: 'TagWriter') -> None:
        for attrib in self.block_attribs:
            tagwriter.write_tag2(330, attrib.handle)
            tagwriter.write_tag2(177, attrib.index)
            tagwriter.write_tag2(44, attrib.width)
            tagwriter.write_tag2(302, attrib.text)


class MultiLeaderContext:
    ATTRIBS = {
        40: 'scale',
        10: 'base_point',
        41: 'text_height',
        140: 'arrowhead_size',
        145: 'landing_gap_size',
        174: 'left_attachment',
        175: 'right_attachment',
        176: 'text_align_type',
        177: 'attachment_type',
        110: 'plane_origin',
        111: 'plane_x_axis',
        112: 'plane_y_axis',
        297: 'plane_normal_reversed',
        272: 'top_attachment',
        273: 'bottom_attachment',
    }

    def __init__(self):
        self.leaders: List['Leader'] = []
        self.scale: float = 1.0  # overall scale
        self.base_point: Vec3 = NULLVEC
        self.text_height = 4.0
        self.arrowhead_size = 4.0
        self.landing_gap_size = 2.0
        self.left_attachment = 1
        self.right_attachment = 1
        self.text_align_type = 0  # 0=left, 1=center, 2=right
        self.attachment_type = 0  # 0=content extents, 1=insertion point
        self.mtext: Optional[MTextData] = None
        self.block: Optional[BlockData] = None
        self.plane_origin: Vec3 = NULLVEC
        self.plane_x_axis: Vec3 = X_AXIS
        self.plane_y_axis: Vec3 = Y_AXIS
        self.plane_normal_reversed: int = 0
        self.top_attachment = 9
        self.bottom_attachment = 9

    @classmethod
    def load(cls, context: List[Union['DXFTag', List]]) -> 'MultiLeaderContext':
        assert context[0] == (START_CONTEXT_DATA, CONTEXT_STR)
        ctx = cls()
        content = None
        for tag in context:
            if isinstance(tag, list):  # Leader()
                ctx.leaders.append(Leader.load(tag))
                continue
            # parse context tags
            code, value = tag
            if content:
                if content.parse(code, value):
                    continue
                else:
                    content = None

            if code == 290 and value == 1:
                content = MTextData()
                ctx.mtext = content
            elif code == 296 and value == 1:
                content = BlockData()
                ctx.block = content
            else:
                name = MultiLeaderContext.ATTRIBS.get(code)
                if name:
                    ctx.__setattr__(name, value)
        return ctx

    @property
    def is_valid(self) -> bool:
        return True

    def export_dxf(self, tagwriter: 'TagWriter') -> None:
        write_tag2 = tagwriter.write_tag2
        write_vertex = tagwriter.write_vertex
        write_tag2(START_CONTEXT_DATA, CONTEXT_STR)
        # All MultiLeaderContext tags:
        write_tag2(40, self.scale)
        write_vertex(10, self.base_point)
        write_tag2(41, self.text_height)
        write_tag2(140, self.arrowhead_size)
        write_tag2(145, self.landing_gap_size)
        write_tag2(174, self.left_attachment)
        write_tag2(175, self.right_attachment)
        write_tag2(176, self.text_align_type)
        write_tag2(177, self.attachment_type)

        if self.mtext:
            write_tag2(290, 1)  # has mtext content
            self.mtext.export_dxf(tagwriter)
        else:
            write_tag2(290, 0)

        if self.block:
            write_tag2(296, 1)  # has block content
            self.block.export_dxf(tagwriter)
        else:
            write_tag2(296, 0)

        write_vertex(110, self.plane_origin)
        write_vertex(111, self.plane_x_axis)
        write_vertex(112, self.plane_y_axis)
        write_tag2(297, self.plane_normal_reversed)

        # Export Leader and LiederLine objects:
        for leader in self.leaders:
            leader.export_dxf(tagwriter)

        # Additional MultiLeaderContext tags:
        if tagwriter.dxfversion >= const.DXF2010:
            write_tag2(272, self.top_attachment)
            write_tag2(273, self.bottom_attachment)
        write_tag2(END_CONTEXT_DATA, '}')


class MTextData:
    ATTRIBS = {
        304: 'default_content',
        11: 'normal_direction',
        340: 'style_handle',
        12: 'location',
        13: 'direction',
        42: 'rotation',
        43: 'boundary_width',
        44: 'boundary_height',
        45: 'line_space_factor',
        170: 'line_space_style',
        90: 'color',
        171: 'alignment',
        172: 'flow_direction',
        91: 'bg_color',
        141: 'bg_scale_factor',
        92: 'bg_transparency',
        291: 'has_bg_color',
        292: 'has_bg_fill',
        173: 'column_type',
        293: 'use_auto_height',
        142: 'column_width',
        143: 'column_gutter_width',
        294: 'column_flow_reversed',
        144: 'column_sizes',  # multiple values
        295: 'use_word_break',
    }

    def __init__(self):
        self.default_content: str = ''
        self.normal_direction: Vec3 = Z_AXIS
        self.style_handle = None  # handle of TextStyle() table entry
        self.location: Vec3 = NULLVEC
        self.direction: Vec3 = X_AXIS  # text direction
        self.rotation: float = 0.  # in radians!
        self.boundary_width: float = 0.
        self.boundary_height: float = 0.
        self.line_space_factor: float = 1.0
        self.line_space_style: int = 1  # 1=at least, 2=exactly
        self.color: int = colors.BY_BLOCK_RAW_VALUE
        self.alignment: int = 1  # 1=left, 2=center, 3=right
        self.flow_direction: int = 1  # 1=horiz, 3=vert, 6=by style
        self.bg_color: int = -939524096  # use window background color? (CMC)
        self.bg_scale_factor: float = 1.5
        self.bg_transparency: int = 0
        self.has_bg_color: int = 0
        self.has_bg_fill: int = 0
        self.column_type: int = 0  # unknown values
        self.use_auto_height: int = 0
        self.column_width: float = 0.
        self.column_gutter_width: float = 0.
        self.column_flow_reversed: int = 0
        self.column_sizes: List[float] = []  # heights?
        self.use_word_break: int = 1

    def parse(self, code: int, value) -> bool:
        # return True if data belongs to mtext else False (end of mtext section)
        if code == 144:
            self.column_sizes.append(value)
            return True
        attrib = MTextData.ATTRIBS.get(code)
        if attrib:
            self.__setattr__(attrib, value)
        return bool(attrib)

    def export_dxf(self, tagwriter: 'TagWriter') -> None:
        write_tag2 = tagwriter.write_tag2
        write_vertex = tagwriter.write_vertex
        write_tag2(304, self.default_content)
        write_vertex(11, self.normal_direction)
        if self.style_handle:
            write_tag2(340, self.style_handle)
        else:
            # Do not write None, but "0" is also not valid!
            # DXF structure error should be detected before export.
            write_tag2(340, "0")
        write_vertex(12, self.location)
        write_vertex(13, self.direction)
        write_tag2(42, self.rotation)
        write_tag2(43, self.boundary_width)
        write_tag2(44, self.boundary_height)
        write_tag2(45, self.line_space_factor)
        write_tag2(170, self.line_space_style)
        write_tag2(90, self.color)
        write_tag2(171, self.alignment)
        write_tag2(172, self.flow_direction)
        write_tag2(91, self.bg_color)
        write_tag2(141, self.bg_scale_factor)
        write_tag2(92, self.bg_transparency)
        write_tag2(291, self.has_bg_color)
        write_tag2(292, self.has_bg_fill)
        write_tag2(173, self.column_type)
        write_tag2(293, self.use_auto_height)
        write_tag2(142, self.column_width)
        write_tag2(143, self.column_gutter_width)
        write_tag2(294, self.column_flow_reversed)
        for size in self.column_sizes:
            write_tag2(144, size)
        write_tag2(295, self.use_word_break)


class BlockData:
    ATTRIBS = {
        341: 'block_record_handle',
        14: 'normal_direction',
        15: 'location',
        16: 'scale',
        46: 'rotation',
        93: 'color',
    }

    def __init__(self):
        self.block_record_handle = None
        self.normal_direction: Vec3 = Z_AXIS
        self.location: Vec3 = NULLVEC
        self.scale: Vec3 = Vec3(1, 1, 1)
        self.rotation: float = 0  # in radians!
        self.color: int = colors.BY_BLOCK_RAW_VALUE
        # The transformation matrix is stored in transposed order
        # of ezdxf.math.Matrix44()!
        self._matrix: List[float] = []  # group code 47 x 16

    @property
    def matrix44(self) -> Matrix44:
        m = Matrix44(self._matrix)
        m.transpose()
        return m

    @matrix44.setter
    def matrix44(self, m: Matrix44) -> None:
        m = m.copy()
        m.transpose()
        self._matrix = list(m)

    def parse(self, code: int, value) -> bool:
        attrib = BlockData.ATTRIBS.get(code)
        if attrib:
            self.__setattr__(attrib, value)
        elif code == 47:
            self._matrix.append(value)
        else:
            return False
        # return True if data belongs to block else False (end of block section)
        return True

    def export_dxf(self, tagwriter: 'TagWriter') -> None:
        write_tag2 = tagwriter.write_tag2
        write_vertex = tagwriter.write_vertex
        if self.block_record_handle:
            write_tag2(341, self.block_record_handle)
        else:
            # Do not write None, but "0" is also not valid!
            # DXF structure error should be detected before export.
            write_tag2(341, "0")
        write_vertex(14, self.normal_direction)
        write_vertex(15, self.location)
        write_vertex(16, self.scale)
        write_tag2(46, self.rotation)
        write_tag2(93, self.color)
        for value in self._matrix:
            write_tag2(47, value)


class Leader:
    def __init__(self):
        self.lines: List['LeaderLine'] = []
        self.has_last_leader_line: int = 0  # group code 290
        self.has_dogleg_vector: int = 0  # group code 291
        self.last_leader_point: Vec3 = NULLVEC  # group code (10, 20, 30)
        self.dogleg_vector: Vec3 = X_AXIS  # group code (11, 21, 31)
        self.dogleg_length: float = 1.0  # group code 40
        self.index: int = 0  # group code 90
        self.attachment_direction: int = 0  # group code 271, R21010+
        self.breaks = []  # group code 12, 13 - multiple breaks possible!

    @classmethod
    def load(cls, context: List[Union['DXFTag', List]]):
        assert context[0] == (START_LEADER, LEADER_STR)
        leader = cls()
        for tag in context:
            if isinstance(tag, list):  # LeaderLine()
                leader.lines.append(LeaderLine.load(tag))
                continue

            code, value = tag
            if code == 290:
                leader.has_last_leader_line = value
            elif code == 291:
                leader.has_dogleg_vector = value
            elif code == 10:
                leader.last_leader_point = value
            elif code == 11:
                leader.dogleg_vector = value
            elif code == 40:
                leader.dogleg_length = value
            elif code == 90:
                leader.index = value
            elif code == 271:
                leader.attachment_direction = value
            elif code in (12, 13):
                leader.breaks.append(value)

        return leader

    def export_dxf(self, tagwriter: 'TagWriter') -> None:
        write_tag2 = tagwriter.write_tag2
        write_vertex = tagwriter.write_vertex

        write_tag2(START_LEADER, LEADER_STR)
        write_tag2(290, self.has_last_leader_line)
        write_tag2(291, self.has_dogleg_vector)
        if self.has_last_leader_line:
            write_vertex(10, self.last_leader_point)
        if self.has_dogleg_vector:
            write_vertex(11, self.dogleg_vector)

        code = 0
        for vertex in self.breaks:
            # write alternate group code 12 and 13
            write_vertex(12 + code, vertex)
            code = 1 - code
        write_tag2(90, self.index)
        write_tag2(40, self.dogleg_length)

        # Export leader lines:
        for line in self.lines:
            line.export_dxf(tagwriter)

        if tagwriter.dxfversion >= const.DXF2010:
            write_tag2(271, self.attachment_direction)
        write_tag2(END_LEADER, '}')


class LeaderLine:
    def __init__(self):
        self.vertices: List[Vec3] = []
        self.breaks: Optional[List[Union[int, Vec3]]] = None
        # Breaks: 90, 11, 12, [11, 12, ...] [, 90, 11, 12 [11, 12, ...]]
        # group code 90 = break index
        # group code 11 = start vertex of break
        # group code 12 = end vertex of break
        # multiple breaks per index possible
        self.index: int = 0  # group code 91
        self.color: int = colors.BY_BLOCK_RAW_VALUE  # group code 92
        # R2010+: override properties see ODA DWG pg. 214-215

    @classmethod
    def load(cls, tags: List['DXFTag']):
        assert tags[0] == (START_LEADER_LINE, LEADER_LINE_STR)
        line = LeaderLine()
        vertices = line.vertices
        breaks = []
        for code, value in tags:
            if code == 10:
                vertices.append(value)
            elif code in (90, 11, 12):
                breaks.append(value)
            elif code == 91:
                line.index = value
            elif code == 92:
                line.color = value
        if breaks:
            line.breaks = breaks
        return line

    def export_dxf(self, tagwriter: 'TagWriter') -> None:
        write_tag2 = tagwriter.write_tag2
        write_vertex = tagwriter.write_vertex

        write_tag2(START_LEADER_LINE, LEADER_LINE_STR)
        for vertex in self.vertices:
            write_vertex(10, vertex)
        if self.breaks:
            code = 0
            for value in self.breaks:
                if isinstance(value, int):
                    # break index
                    write_tag2(90, value)
                else:
                    # 11 .. start vertex of break
                    # 12 .. end vertex of break
                    write_vertex(11 + code, value)
                    code = 1 - code
        write_tag2(91, self.index)
        write_tag2(92, self.color)
        write_tag2(END_LEADER_LINE, '}')


@register_entity
class MLeader(MultiLeader):
    DXFTYPE = 'MLEADER'


acdb_mleader_style = DefSubclass('AcDbMLeaderStyle', {
    'unknown1': DXFAttr(179, default=2),
    'content_type': DXFAttr(170, default=2),
    'draw_mleader_order_type': DXFAttr(171, default=1),
    'draw_leader_order_type': DXFAttr(172, default=0),
    'max_leader_segments_points': DXFAttr(90, default=2),
    'first_segment_angle_constraint': DXFAttr(40, default=0.0),
    'second_segment_angle_constraint': DXFAttr(41, default=0.0),
    'leader_type': DXFAttr(173, default=1),
    'leader_line_color': DXFAttr(91, default=-1056964608),
    # raw color: BY_BLOCK
    # raw color: BY_BLOCK
    'leader_linetype_handle': DXFAttr(340),
    'leader_lineweight': DXFAttr(92, default=-2),
    'has_landing': DXFAttr(290, default=1),
    'landing_gap': DXFAttr(42, default=2.0),
    'has_dogleg': DXFAttr(291, default=1),
    'dogleg_length': DXFAttr(43, default=8),
    'name': DXFAttr(3, default='Standard'),

    # no handle is default arrow 'closed filled':
    'arrow_head_handle': DXFAttr(341),
    'arrow_head_size': DXFAttr(44, default=4),
    'default_text_content': DXFAttr(300, default=''),
    'text_style_handle': DXFAttr(342),
    'text_left_attachment_type': DXFAttr(174, default=1),
    'text_angle_type': DXFAttr(175, default=1),
    'text_alignment_type': DXFAttr(176, default=0),
    'text_right_attachment_type': DXFAttr(178, default=1),
    'text_color': DXFAttr(93, default=-1056964608),  # raw color: BY_BLOCK
    'text_height': DXFAttr(45, default=4),
    'has_frame_text': DXFAttr(292, default=0),
    'text_align_always_left': DXFAttr(297, default=0),
    'align_space': DXFAttr(46, default=4),
    'has_block_scaling': DXFAttr(293),
    'block_record_handle': DXFAttr(343),
    'block_color': DXFAttr(94, default=-1056964608),  # raw color: BY_BLOCK
    'block_scale_x': DXFAttr(47, default=1),
    'block_scale_y': DXFAttr(49, default=1),
    'block_scale_z': DXFAttr(140, default=1),
    'has_block_rotation': DXFAttr(294, default=1),
    'block_rotation': DXFAttr(141, default=0),
    'block_connection_type': DXFAttr(177, default=0),
    'scale': DXFAttr(142, default=1),
    'overwrite_property_value': DXFAttr(295, default=0),
    'is_annotative': DXFAttr(296, default=0),
    'break_gap_size': DXFAttr(143, default=3.75),

    # 0 = Horizontal; 1 = Vertical:
    'text_attachment_direction': DXFAttr(271, default=0),

    # 9 = Center; 10 = Underline and Center:
    'text_bottom__attachment_direction': DXFAttr(272, default=9),

    # 9 = Center; 10 = Overline and Center:
    'text_top_attachment_direction': DXFAttr(273, default=9),

    'unknown2': DXFAttr(298, optional=True),  # boolean flag ?
})
acdb_mleader_style_group_codes = group_code_mapping(acdb_mleader_style)


@register_entity
class MLeaderStyle(DXFObject):
    DXFTYPE = 'MLEADERSTYLE'
    DXFATTRIBS = DXFAttributes(base_class, acdb_mleader_style)
    MIN_DXF_VERSION_FOR_EXPORT = const.DXF2000

    def load_dxf_attribs(
            self, processor: SubclassProcessor = None) -> 'DXFNamespace':
        dxf = super().load_dxf_attribs(processor)
        if processor:
            processor.fast_load_dxfattribs(
                dxf, acdb_mleader_style_group_codes, subclass=1)
        return dxf

    def export_entity(self, tagwriter: 'TagWriter') -> None:
        super().export_entity(tagwriter)
        tagwriter.write_tag2(const.SUBCLASS_MARKER, acdb_mleader_style.name)
        self.dxf.export_dxf_attribs(
            tagwriter, acdb_mleader_style.attribs.keys())


class MLeaderStyleCollection(ObjectCollection):
    def __init__(self, doc: 'Drawing'):
        super().__init__(
            doc, dict_name='ACAD_MLEADERSTYLE', object_type='MLEADERSTYLE')
        self.create_required_entries()

    def create_required_entries(self) -> None:
        for name in ('Standard',):
            if name not in self.object_dict:
                mleader_style = self.new(name)
                # set standard text style
                text_style = self.doc.styles.get('Standard')
                mleader_style.dxf.text_style_handle = text_style.dxf.handle



# Purpose: curve objects
# Created: 26.03.2010, 2018 adapted for ezdxf
# Copyright (c) 2010-2018, Manfred Moitzi
# License: MIT License
from typing import TYPE_CHECKING, Iterable, List, Tuple, Optional
import random
import math
from ezdxf.math import Vec3, Vec2, Matrix44, perlin
from ezdxf.math.bspline import global_bspline_interpolation
from ezdxf.math.bspline import BSpline, BSplineU, BSplineClosed
from ezdxf.math.bezier4p import Bezier4P
from ezdxf.math.eulerspiral import EulerSpiral as _EulerSpiral

if TYPE_CHECKING:
    from ezdxf.eztypes import Vertex, BaseLayout, Matrix44


def rnd(max_value):
    return max_value / 2.0 - random.random() * max_value


def rnd_perlin(max_value, walker):
    r = perlin.snoise2(walker.x, walker.y)
    return max_value / 2.0 - r * max_value


def random_2d_path(steps: int = 100, max_step_size: float = 1.0, max_heading: float = math.pi / 2,
                   retarget: int = 20) -> Iterable[Vec2]:
    """
    Returns a random 2D path as iterable of :class:`~ezdxf.math.Vec2` objects.

    Args:
        steps: count of vertices to generate
        max_step_size: max step size
        max_heading: limit heading angle change per step to ± max_heading/2 in radians
        retarget: specifies steps before changing global walking target

    """
    max_ = max_step_size * steps

    def next_global_target():
        return Vec2((rnd(max_), rnd(max_)))

    walker = Vec2(0, 0)
    target = next_global_target()
    for i in range(steps):
        if i % retarget == 0:
            target = target + next_global_target()
        angle = (target - walker).angle
        heading = angle + rnd_perlin(max_heading, walker)
        length = max_step_size * random.random()
        walker = walker + Vec2.from_angle(heading, length)
        yield walker


def random_3d_path(steps: int = 100, max_step_size: float = 1.0, max_heading: float = math.pi / 2.0,
                   max_pitch: float = math.pi / 8.0, retarget: int = 20) -> Iterable[Vec3]:
    """
    Returns a random 3D path as iterable of :class:`~ezdxf.math.Vec3` objects.

    Args:
        steps: count of vertices to generate
        max_step_size: max step size
        max_heading: limit heading angle change per step to ± max_heading/2, rotation about the z-axis in radians
        max_pitch: limit pitch angle change per step to ± max_pitch/2, rotation about the x-axis in radians
        retarget: specifies steps before changing global walking target

    """
    max_ = max_step_size * steps

    def next_global_target():
        return Vec3((rnd(max_), rnd(max_), rnd(max_)))

    walker = Vec3()
    target = next_global_target()
    for i in range(steps):
        if i % retarget == 0:
            target = target + next_global_target()
        angle = (target - walker).angle
        length = max_step_size * random.random()
        heading_angle = angle + rnd_perlin(max_heading, walker)
        next_step = Vec3.from_angle(heading_angle, length)
        pitch_angle = rnd_perlin(max_pitch, walker)
        walker += Matrix44.x_rotate(pitch_angle).transform(next_step)
        yield walker


class Bezier:
    """
    Bezier 2D/3D curve.

    The Bezier() class is implemented with multiple segments, each segment is an optimized 4 point bezier curve, the
    4 control points of the curve are: the start point (1) and the end point (4), point (2) is start point + start vector
    and point (3) is end point + end vector. Each segment has its own approximation count.

    """

    class Segment:
        def __init__(self, start: 'Vertex', end: 'Vertex', start_tangent: 'Vertex', end_tangent: 'Vertex',
                     segments: int):
            self.start = Vec3(start)
            self.end = Vec3(end)
            self.start_tangent = Vec3(start_tangent)  # as vector, from start point
            self.end_tangent = Vec3(end_tangent)  # as vector, from end point
            self.segments = segments

        def approximate(self) -> Iterable[Vec3]:
            control_points = [
                self.start,
                self.start + self.start_tangent,
                self.end + self.end_tangent,
                self.end,
            ]
            bezier = Bezier4P(control_points)
            return bezier.approximate(self.segments)

    def __init__(self):
        # fit point, first control vector, second control vector, segment count
        self.points = []  # type: List[Tuple[Vec3, Optional[Vec3], Optional[Vec3], Optional[int]]]

    def start(self, point: 'Vertex', tangent: 'Vertex') -> None:
        """
        Set start point and start tangent.

        Args:
            point: start point as :class:`~ezdxf.math.Vec3` or ``(x, y, z)`` tuple
            tangent: start tangent as vector, example: ``(5, 0, 0)`` means a
                     horizontal tangent with a length of 5 drawing units
        """
        self.points.append((point, None, tangent, None))

    def append(self, point: 'Vertex', tangent1: 'Vertex', tangent2: 'Vertex' = None, segments: int = 20):
        """
        Append a control point with two control tangents.

        Args:
            point: control point as :class:`~ezdxf.math.Vec3` or ``(x, y, z)`` tuple
            tangent1: first control tangent as vector "left" of control point
            tangent2: second control tangent as vector "right" of control point, if omitted `tangent2` = `-tangent1`
            segments: count of line segments for polyline approximation, count of line segments from previous
                      control point to appended control point.

        """
        tangent1 = Vec3(tangent1)
        if tangent2 is None:
            tangent2 = -tangent1
        else:
            tangent2 = Vec3(tangent2)
        self.points.append((point, tangent1, tangent2, int(segments)))

    def _build_bezier_segments(self) -> Iterable[Segment]:
        if len(self.points) > 1:
            for from_point, to_point in zip(self.points[:-1], self.points[1:]):
                start_point = from_point[0]
                start_tangent = from_point[2]  # tangent2
                end_point = to_point[0]
                end_tangent = to_point[1]  # tangent1
                count = to_point[3]
                yield Bezier.Segment(start_point, end_point,
                                     start_tangent, end_tangent, count)
        else:
            raise ValueError('Two or more points needed!')

    def render(self, layout: 'BaseLayout', force3d: bool = False, dxfattribs: dict = None) -> None:
        """
        Render bezier curve as 2D/3D :class:`~ezdxf.entities.Polyline`.

        Args:
            layout: :class:`~ezdxf.layouts.BaseLayout` object
            force3d: force 3D polyline rendering
            dxfattribs: DXF attributes for :class:`~ezdxf.entities.Polyline`

        """
        points = []
        for segment in self._build_bezier_segments():
            points.extend(segment.approximate())
        if force3d or any(p[2] for p in points):
            layout.add_polyline3d(points, dxfattribs=dxfattribs)
        else:
            layout.add_polyline2d(points, dxfattribs=dxfattribs)


class Spline:
    def __init__(self, points: Iterable['Vertex'] = None, segments: int = 100):
        """
        Args:
            points: spline definition points as :class:`~ezdxf.math.Vec3` or ``(x, y, z)`` tuple
            segments: count of line segments for approximation, vertex count is `segments` + 1
        """
        if points is None:
            points = []
        self.points = points
        self.segments = int(segments)

    def subdivide(self, segments: int = 4) -> None:
        """
        Calculate overall segment count, where segments is the sub-segment count, `segments` = 4, means 4 line
        segments between two definition points e.g. 4 definition points and 4 segments = 12 overall segments, useful
        for fit point rendering.

        Args:
            segments: sub-segments count between two definition points

        """
        self.segments = (len(self.points) - 1) * segments

    def render_as_fit_points(self, layout: 'BaseLayout', degree: int = 3, method: str = 'chord',
                             dxfattribs: dict = None) -> None:
        """
        Render a B-spline as 2D/3D :class:`~ezdxf.entities.Polyline`, where the definition points are fit points.

           - 2D spline vertices uses: :meth:`~ezdxf.layouts.BaseLayout.add_polyline2d`
           - 3D spline vertices uses: :meth:`~ezdxf.layouts.BaseLayout.add_polyline3d`

        Args:
            layout: :class:`~ezdxf.layouts.BaseLayout` object
            degree: degree of B-spline (order = `degree` + 1)
            method: "uniform", "distance"/"chord", "centripetal"/"sqrt_chord" or "arc"
                    calculation method for parameter t
            dxfattribs: DXF attributes for :class:`~ezdxf.entities.Polyline`

        """
        spline = global_bspline_interpolation(self.points, degree=degree, method=method)
        vertices = list(spline.approximate(self.segments))
        if any(vertex.z != 0. for vertex in vertices):
            layout.add_polyline3d(vertices, dxfattribs=dxfattribs)
        else:
            layout.add_polyline2d(vertices, dxfattribs=dxfattribs)

    render = render_as_fit_points

    def render_open_bspline(self, layout: 'BaseLayout', degree: int = 3, dxfattribs: dict = None) -> None:
        """
        Render an open uniform BSpline as 3D :class:`~ezdxf.entities.Polyline`. Definition points are control points.

        Args:
            layout: :class:`~ezdxf.layouts.BaseLayout` object
            degree: degree of B-spline (order = `degree` + 1)
            dxfattribs: DXF attributes for :class:`~ezdxf.entities.Polyline`

        """
        spline = BSpline(self.points, order=degree + 1)
        layout.add_polyline3d(list(spline.approximate(self.segments)), dxfattribs=dxfattribs)

    def render_uniform_bspline(self, layout: 'BaseLayout', degree: int = 3, dxfattribs: dict = None) -> None:
        """
        Render a uniform BSpline as 3D :class:`~ezdxf.entities.Polyline`. Definition points are control points.

        Args:
            layout: :class:`~ezdxf.layouts.BaseLayout` object
            degree: degree of B-spline (order = `degree` + 1)
            dxfattribs: DXF attributes for :class:`~ezdxf.entities.Polyline`

        """
        spline = BSplineU(self.points, order=degree + 1)
        layout.add_polyline3d(list(spline.approximate(self.segments)), dxfattribs=dxfattribs)

    def render_closed_bspline(self, layout: 'BaseLayout', degree: int = 3, dxfattribs: dict = None) -> None:
        """
        Render a closed uniform BSpline as 3D :class:`~ezdxf.entities.Polyline`. Definition points are control points.

        Args:
            layout: :class:`~ezdxf.layouts.BaseLayout` object
            degree: degree of B-spline (order = `degree` + 1)
            dxfattribs: DXF attributes for :class:`~ezdxf.entities.Polyline`

        """
        spline = BSplineClosed(self.points, order=degree + 1)
        layout.add_polyline3d(list(spline.approximate(self.segments)), dxfattribs=dxfattribs)

    def render_open_rbspline(self, layout: 'BaseLayout', weights: Iterable[float], degree: int = 3,
                             dxfattribs: dict = None) -> None:
        """
        Render a rational open uniform BSpline as 3D :class:`~ezdxf.entities.Polyline`. Definition points are control
        points.

        Args:
            layout: :class:`~ezdxf.layouts.BaseLayout` object
            weights: list of weights, requires a weight value (float) for each definition point.
            degree: degree of B-spline (order = `degree` + 1)
            dxfattribs: DXF attributes for :class:`~ezdxf.entities.Polyline`

        """
        spline = BSpline(self.points, order=degree + 1, weights=weights)
        layout.add_polyline3d(list(spline.approximate(self.segments)), dxfattribs=dxfattribs)

    def render_uniform_rbspline(self, layout: 'BaseLayout', weights: Iterable[float], degree: int = 3,
                                dxfattribs: dict = None) -> None:
        """
        Render a rational uniform BSpline as 3D :class:`~ezdxf.entities.Polyline`. Definition points are control
        points.

        Args:
            layout: :class:`~ezdxf.layouts.BaseLayout` object
            weights: list of weights, requires a weight value (float) for each definition point.
            degree: degree of B-spline (order = `degree` + 1)
            dxfattribs: DXF attributes for :class:`~ezdxf.entities.Polyline`

        """
        spline = BSplineU(self.points, order=degree + 1, weights=weights)
        layout.add_polyline3d(list(spline.approximate(self.segments)), dxfattribs=dxfattribs)

    def render_closed_rbspline(self, layout: 'BaseLayout', weights: Iterable[float], degree: int = 3,
                               dxfattribs: dict = None) -> None:
        """
        Render a rational BSpline as 3D :class:`~ezdxf.entities.Polyline`. Definition points are control
        points.

        Args:
            layout: :class:`~ezdxf.layouts.BaseLayout` object
            weights: list of weights, requires a weight value (float) for each definition point.
            degree: degree of B-spline (order = `degree` + 1)
            dxfattribs: DXF attributes for :class:`~ezdxf.entities.Polyline`

        """
        spline = BSplineClosed(self.points, order=degree + 1, weights=weights)
        layout.add_polyline3d(list(spline.approximate(self.segments)), dxfattribs=dxfattribs)


class EulerSpiral:
    """
    Euler spiral (clothoid) for `curvature` (Radius of curvature).

    This is a parametric curve, which always starts at the origin (0, 0).

    """

    def __init__(self, curvature: float = 1):
        """
        Args:
            curvature: Radius of curvature

        """
        self.spiral = _EulerSpiral(float(curvature))

    def render_polyline(self, layout: 'BaseLayout', length: float = 1, segments: int = 100,
                        matrix: 'Matrix44' = None, dxfattribs: dict = None):
        """
        Render curve as :class:`~ezdxf.entities.Polyline`.

        Args:
            layout: :class:`~ezdxf.layouts.BaseLayout` object
            length: length measured along the spiral curve from its initial position
            segments: count of line segments to use, vertex count is `segments` + 1
            matrix: transformation matrix as :class:`~ezdxf.math.Matrix44`
            dxfattribs: DXF attributes for :class:`~ezdxf.entities.Polyline`

        Returns:
            :class:`~ezdxf.entities.Polyline`

        """
        points = self.spiral.approximate(length, segments)
        if matrix is not None:
            points = matrix.transform_vertices(points)
        return layout.add_polyline3d(list(points), dxfattribs=dxfattribs)

    def render_spline(self, layout: 'BaseLayout', length: float = 1, fit_points: int = 10, degree: int = 3,
                      matrix: 'Matrix44' = None, dxfattribs: dict = None):
        """
        Render curve as :class:`~ezdxf.entities.Spline`.

        Args:
            layout: :class:`~ezdxf.layouts.BaseLayout` object
            length: length measured along the spiral curve from its initial position
            fit_points: count of spline fit points to use
            degree: degree of B-spline
            matrix: transformation matrix as :class:`~ezdxf.math.Matrix44`
            dxfattribs: DXF attributes for :class:`~ezdxf.entities.Spline`

        Returns:
            :class:`~ezdxf.entities.Spline`

        """
        spline = self.spiral.bspline(length, fit_points, degree=degree)
        points = spline.control_points
        if matrix is not None:
            points = matrix.transform_vertices(points)
        return layout.add_open_spline(
            control_points=points,
            degree=spline.degree,
            knots=spline.knots(),
            dxfattribs=dxfattribs,
        )



# Copyright (c) 2019-2020 Manfred Moitzi
# License: MIT License
from typing import TYPE_CHECKING, Tuple
from .dxfentity import DXFEntity, SubclassProcessor, DXFNamespace
from ezdxf.lldxf.attributes import (
    DXFAttr, DXFAttributes, DefSubclass, group_code_mapping,
)
from ezdxf.lldxf.const import DXF2004, DXF2000
from .factory import register_entity

if TYPE_CHECKING:
    from ezdxf.eztypes import Drawing, ExtendedTags, TagWriter

__all__ = ['DXFClass']

class_def = DefSubclass(None, {
    # Class DXF record name; always unique
    'name': DXFAttr(1),
    # C++ class name. Used to bind with software that defines object class
    # behavior; always unique
    'cpp_class_name': DXFAttr(2),
    # Application name. Posted in Alert box when a class definition listed in
    # this section is not currently loaded
    'app_name': DXFAttr(3),
    # Proxy capabilities flag. Bit-coded value that indicates the capabilities
    # of this object as a proxy:
    # 0 = No operations allowed (0)
    # 1 = Erase allowed (0x1)
    # 2 = Transform allowed (0x2)
    # 4 = Color change allowed (0x4)
    # 8 = Layer change allowed (0x8)
    # 16 = Linetype change allowed (0x10)
    # 32 = Linetype scale change allowed (0x20)
    # 64 = Visibility change allowed (0x40)
    # 128 = Cloning allowed (0x80)
    # 256 = Lineweight change allowed (0x100)
    # 512 = Plot Style Name change allowed (0x200)
    # 895 = All operations except cloning allowed (0x37F)
    # 1023 = All operations allowed (0x3FF)
    # 1024 = Disables proxy warning dialog (0x400)
    # 32768 = R13 format proxy (0x8000)
    'flags': DXFAttr(90, default=0),
    # Instance count for a custom class
    'instance_count': DXFAttr(91, dxfversion=DXF2004, default=0),
    # Was-a-proxy flag. Set to 1 if class was not loaded when this DXF file was
    # created, and 0 otherwise
    'was_a_proxy': DXFAttr(280, default=0),
    # Is-an-entity flag. Set to 1 if class was derived from the AcDbEntity class
    # and can reside in the BLOCKS or ENTITIES section. If 0, instances may
    # appear only in the OBJECTS section
    'is_an_entity': DXFAttr(281, default=0),
})
class_def_group_codes = group_code_mapping(class_def)


@register_entity
class DXFClass(DXFEntity):
    DXFTYPE = 'CLASS'
    DXFATTRIBS = DXFAttributes(class_def)
    MIN_DXF_VERSION_FOR_EXPORT = DXF2000

    @classmethod
    def new(cls, handle: str = None, owner: str = None, dxfattribs: dict = None,
            doc: 'Drawing' = None) -> 'DXFClass':
        """ New CLASS constructor - has no handle, no owner and do not need
        document reference .
        """
        dxf_class = cls()
        dxf_class.doc = doc
        dxfattribs = dxfattribs or {}
        dxf_class.update_dxf_attribs(dxfattribs)
        return dxf_class

    def load_tags(self, tags: 'ExtendedTags', dxfversion=None) -> None:
        """ Called by load constructor. CLASS is special. """
        if tags:
            # do not process base class!!!
            self.dxf = DXFNamespace(entity=self)
            processor = SubclassProcessor(tags)
            processor.fast_load_dxfattribs(
                self.dxf, class_def_group_codes, 0, log=False)

    def export_dxf(self, tagwriter: 'TagWriter'):
        """ Do complete export here, because CLASS is special. """
        dxfversion = tagwriter.dxfversion
        if dxfversion < DXF2000:
            return
        attribs = self.dxf
        tagwriter.write_tag2(0, self.DXFTYPE)
        attribs.export_dxf_attribs(tagwriter, [
            'name', 'cpp_class_name', 'app_name', 'flags', 'instance_count',
            'was_a_proxy', 'is_an_entity',
        ])

    @property
    def key(self) -> Tuple[str, str]:
        return self.dxf.name, self.dxf.cpp_class_name



# Copyright (c) 2019-2020 Manfred Moitzi
# License: MIT License
from typing import TYPE_CHECKING, List
from ezdxf.lldxf import validator
from ezdxf.lldxf.attributes import (
    DXFAttr, DXFAttributes, DefSubclass, XType, RETURN_DEFAULT,
    group_code_mapping,
)
from ezdxf.lldxf.const import DXF12, SUBCLASS_MARKER, VERTEXNAMES
from ezdxf.math import Matrix44, Z_AXIS, NULLVEC, Vec3
from ezdxf.math.transformtools import OCSTransform
from .dxfentity import base_class, SubclassProcessor
from .dxfgfx import DXFGraphic, acdb_entity, elevation_to_z_axis
from .factory import register_entity

if TYPE_CHECKING:
    from ezdxf.eztypes import TagWriter, DXFNamespace

__all__ = ['Solid', 'Trace', 'Face3d']

acdb_trace = DefSubclass('AcDbTrace', {
    # 1. corner Solid WCS; Trace OCS
    'vtx0': DXFAttr(10, xtype=XType.point3d, default=NULLVEC),

    # 2. corner Solid WCS; Trace OCS
    'vtx1': DXFAttr(11, xtype=XType.point3d, default=NULLVEC),

    # 3. corner Solid WCS; Trace OCS
    'vtx2': DXFAttr(12, xtype=XType.point3d, default=NULLVEC),

    # 4. corner Solid WCS; Trace OCS:
    # If only three corners are entered to define the SOLID, then the fourth
    # corner coordinate is the same as the third.
    'vtx3': DXFAttr(13, xtype=XType.point3d, default=NULLVEC),

    # Elevation is a legacy feature from R11 and prior, do not use this
    # attribute, store the entity elevation in the z-axis of the vertices.
    # ezdxf does not export the elevation attribute!
    'elevation': DXFAttr(38, default=0, optional=True),

    # Thickness could be negative:
    'thickness': DXFAttr(39, default=0, optional=True),
    'extrusion': DXFAttr(
        210, xtype=XType.point3d, default=Z_AXIS, optional=True,
        validator=validator.is_not_null_vector,
        fixer=RETURN_DEFAULT,
    ),
})
acdb_trace_group_codes = group_code_mapping(acdb_trace)


class _Base(DXFGraphic):
    def __getitem__(self, num):
        return self.dxf.get(VERTEXNAMES[num])

    def __setitem__(self, num, value):
        return self.dxf.set(VERTEXNAMES[num], value)


@register_entity
class Solid(_Base):
    """ DXF SHAPE entity """
    DXFTYPE = 'SOLID'
    DXFATTRIBS = DXFAttributes(base_class, acdb_entity, acdb_trace)

    def load_dxf_attribs(
            self, processor: SubclassProcessor = None) -> 'DXFNamespace':
        """ Loading interface. (internal API) """
        dxf = super().load_dxf_attribs(processor)
        if processor:
            processor.fast_load_dxfattribs(
                dxf, acdb_trace_group_codes, subclass=2, recover=True)
            if processor.r12:
                # Transform elevation attribute from R11 to z-axis values:
                elevation_to_z_axis(dxf, VERTEXNAMES)
        return dxf

    def export_entity(self, tagwriter: 'TagWriter') -> None:
        """ Export entity specific data as DXF tags. (internal API) """
        super().export_entity(tagwriter)
        if tagwriter.dxfversion > DXF12:
            tagwriter.write_tag2(SUBCLASS_MARKER, acdb_trace.name)
        if not self.dxf.hasattr('vtx3'):
            self.dxf.vtx3 = self.dxf.vtx2
        self.dxf.export_dxf_attribs(tagwriter, [
            'vtx0', 'vtx1', 'vtx2', 'vtx3', 'thickness',
            'extrusion',
        ])

    def transform(self, m: Matrix44) -> 'Solid':
        """ Transform SOLID/TRACE  entity by transformation matrix `m` inplace.

        .. versionadded:: 0.13

        """
        # SOLID/TRACE is 2d entity, placed by an OCS in 3d space
        dxf = self.dxf
        ocs = OCSTransform(self.dxf.extrusion, m)
        for name in VERTEXNAMES:
            if dxf.hasattr(name):
                dxf.set(name, ocs.transform_vertex(dxf.get(name)))
        if dxf.hasattr('thickness'):
            dxf.thickness = ocs.transform_length(
                (0, 0, dxf.thickness), reflection=dxf.thickness)
        dxf.extrusion = ocs.new_extrusion
        return self

    def wcs_vertices(self, close: bool = False) -> List[Vec3]:
        """ Returns WCS vertices in correct order,
        if argument `close` is ``True``, last vertex == first vertex.
        Does **not** return duplicated last vertex if represents a triangle.

        .. versionadded:: 0.15

        """
        ocs = self.ocs()
        return list(ocs.points_to_wcs(self.vertices(close)))

    def vertices(self, close: bool = False) -> List[Vec3]:
        """ Returns OCS vertices in correct order,
        if argument `close` is ``True``, last vertex == first vertex.
        Does **not** return duplicated last vertex if represents a triangle.

        .. versionadded:: 0.15

        """
        dxf = self.dxf
        vertices = [dxf.vtx0, dxf.vtx1, dxf.vtx2]
        if dxf.vtx3 != dxf.vtx2:  # when the face is a triangle, vtx2 == vtx3
            vertices.append(dxf.vtx3)

        # adjust weird vertex order of SOLID and TRACE:
        # 0, 1, 2, 3 -> 0, 1, 3, 2
        if len(vertices) > 3:
            vertices[2], vertices[3] = vertices[3], vertices[2]

        if close and not vertices[0].isclose(vertices[-1]):
            vertices.append(vertices[0])
        return vertices


@register_entity
class Trace(Solid):
    """ DXF TRACE entity """
    DXFTYPE = 'TRACE'


acdb_face = DefSubclass('AcDbFace', {
    # 1. corner WCS:
    'vtx0': DXFAttr(10, xtype=XType.point3d, default=NULLVEC),

    # 2. corner WCS:
    'vtx1': DXFAttr(11, xtype=XType.point3d, default=NULLVEC),

    # 3. corner WCS:
    'vtx2': DXFAttr(12, xtype=XType.point3d, default=NULLVEC),

    # 4. corner WCS:
    # If only three corners are entered to define the SOLID, then the fourth
    # corner coordinate is the same as the third.
    'vtx3': DXFAttr(13, xtype=XType.point3d, default=NULLVEC),

    # invisible:
    # 1 = First edge is invisible
    # 2 = Second edge is invisible
    # 4 = Third edge is invisible
    # 8 = Fourth edge is invisible
    'invisible': DXFAttr(70, default=0, optional=True),
})
acdb_face_group_codes = group_code_mapping(acdb_face)


@register_entity
class Face3d(_Base):
    """ DXF 3DFACE entity """
    DXFTYPE = '3DFACE'
    DXFATTRIBS = DXFAttributes(base_class, acdb_entity, acdb_face)

    def is_invisible_edge(self, num) -> bool:
        """ Returns True if edge `num` is an invisible edge. """
        return bool(self.dxf.invisible & (1 << num))

    def set_edge_visibility(self, num, status=False):
        """ Set visibility of edge `num`, status `True` for visible, status
        `False` for invisible.
        """
        if not status:
            self.dxf.invisible = self.dxf.invisible | (1 << num)
        else:
            self.dxf.invisible = self.dxf.invisible & ~(1 << num)

    def load_dxf_attribs(self,
                         processor: SubclassProcessor = None) -> 'DXFNamespace':
        dxf = super().load_dxf_attribs(processor)
        if processor:
            processor.fast_load_dxfattribs(
                dxf, acdb_face_group_codes, subclass=2, recover=True)
        return dxf

    def export_entity(self, tagwriter: 'TagWriter') -> None:
        super().export_entity(tagwriter)
        if tagwriter.dxfversion > DXF12:
            tagwriter.write_tag2(SUBCLASS_MARKER, acdb_face.name)
        if not self.dxf.hasattr('vtx3'):
            self.dxf.vtx3 = self.dxf.vtx2
        self.dxf.export_dxf_attribs(tagwriter, [
            'vtx0', 'vtx1', 'vtx2', 'vtx3', 'invisible'
        ])

    def transform(self, m: Matrix44) -> 'Face3d':
        """ Transform 3DFACE  entity by transformation matrix `m` inplace.

        .. versionadded:: 0.13

        """
        dxf = self.dxf
        # 3DFACE is a real 3d entity
        dxf.vtx0, dxf.vtx1, dxf.vtx2, dxf.vtx3 = m.transform_vertices(
            (dxf.vtx0, dxf.vtx1, dxf.vtx2, dxf.vtx3))
        return self

    def wcs_vertices(self, close: bool = False) -> List[Vec3]:
        """ Returns WCS vertices, if argument `close` is
        ``True``, last vertex == first vertex.
        Does **not** return duplicated last vertex if represents a triangle.

        Compatibility interface to SOLID and TRACE, 3DFACE vertices are
        already WCS vertices.

        .. versionadded:: 0.15

        """
        dxf = self.dxf
        vertices = [dxf.vtx0, dxf.vtx1, dxf.vtx2]
        if dxf.vtx3 != dxf.vtx2:  # when the face is a triangle, vtx2 == vtx3
            vertices.append(dxf.vtx3)

        if close and not vertices[0].isclose(vertices[-1]):
            vertices.append(vertices[0])
        return vertices



# Copyright (c) 2019-2020, Manfred Moitzi
# License: MIT-License
from typing import TYPE_CHECKING, Iterable, cast, Union, List, Set
from contextlib import contextmanager
import logging
from ezdxf.lldxf import validator
from ezdxf.lldxf.const import DXFValueError, SUBCLASS_MARKER, DXFTypeError
from ezdxf.lldxf.attributes import (
    DXFAttr, DXFAttributes, DefSubclass, RETURN_DEFAULT, group_code_mapping
)
from ezdxf.audit import AuditError
from .dxfentity import base_class, SubclassProcessor, DXFEntity
from .dxfobj import DXFObject
from .factory import register_entity
from .objectcollection import ObjectCollection

logger = logging.getLogger('ezdxf')

if TYPE_CHECKING:
    from ezdxf.eztypes import (
        TagWriter, Drawing, DXFNamespace, Auditor, EntityDB,
    )

__all__ = ['DXFGroup', 'GroupCollection']

acdb_group = DefSubclass('AcDbGroup', {
    # Group description
    'description': DXFAttr(300, default=''),

    # 1 = Unnamed
    # 0 = Named
    'unnamed': DXFAttr(
        70, default=1, validator=validator.is_integer_bool,
        fixer=RETURN_DEFAULT,
    ),

    # 1 = Selectable
    # 0 = Not selectable
    'selectable': DXFAttr(
        71, default=1,
        validator=validator.is_integer_bool,
        fixer=RETURN_DEFAULT,
    ),

    # 340: Hard-pointer handle to entity in group (one entry per object)
})
acdb_group_group_codes = group_code_mapping(acdb_group)
GROUP_ITEM_CODE = 340


@register_entity
class DXFGroup(DXFObject):
    """ Groups are not allowed in block definitions, and each entity can only
    reside in one group, so cloning of groups creates also new entities.

    """
    DXFTYPE = 'GROUP'
    DXFATTRIBS = DXFAttributes(base_class, acdb_group)

    def __init__(self):
        super().__init__()
        self._handles: Set[str] = set()
        self._data: List[DXFEntity] = []

    def copy(self):
        raise DXFTypeError('Copying of GROUP not supported.')

    def load_dxf_attribs(self,
                         processor: SubclassProcessor = None) -> 'DXFNamespace':
        dxf = super().load_dxf_attribs(processor)
        if processor:
            tags = processor.fast_load_dxfattribs(
                dxf, acdb_group_group_codes, 1, log=False)
            self.load_group(tags)
        return dxf

    def load_group(self, tags):
        for code, value in tags:
            if code == GROUP_ITEM_CODE:
                # First store handles, because at this point, objects
                # are not stored in the EntityDB:
                self._handles.add(value)

    def export_entity(self, tagwriter: 'TagWriter') -> None:
        """ Export entity specific data as DXF tags. """
        super().export_entity(tagwriter)
        tagwriter.write_tag2(SUBCLASS_MARKER, acdb_group.name)
        self.dxf.export_dxf_attribs(tagwriter, [
            'description', 'unnamed', 'selectable'])
        self.export_group(tagwriter)

    def export_group(self, tagwriter: 'TagWriter'):
        for entity in self._data:
            if isinstance(entity, str):
                handle = entity
            else:
                handle = entity.dxf.handle
            tagwriter.write_tag2(GROUP_ITEM_CODE, handle)

    def __iter__(self) -> Iterable[DXFEntity]:
        """ Iterate over all DXF entities in :class:`DXFGroup` as instances of
        :class:`DXFGraphic` or inherited (LINE, CIRCLE, ...).

        """
        return (e for e in self._data if e.is_alive)

    def __len__(self) -> int:
        """ Returns the count of DXF entities in :class:`DXFGroup`. """
        return len(self._data)

    def __getitem__(self, item):
        """ Returns entities by standard Python indexing and slicing. """
        return self._data[item]

    def __contains__(self, item: Union[str, DXFEntity]) -> bool:
        """ Returns ``True`` if item is in :class:`DXFGroup`. `item` has to be
        a handle string or an object of type :class:`DXFEntity` or inherited.

        """
        handle = item if isinstance(item, str) else item.dxf.handle
        return handle in set(self.handles())

    def handles(self) -> Iterable[str]:
        """ Iterable of handles of all DXF entities in :class:`DXFGroup`. """
        return (entity.dxf.handle for entity in self)

    def post_load_hook(self, doc: 'Drawing') -> None:
        super().post_load_hook(doc)
        db = doc.entitydb

        def entities():
            for handle in self._handles:
                entity = db.get(handle)
                if entity and entity.is_alive:
                    yield entity

        if len(self._handles):
            self.set_data(entities())
            self._handles = set()

    @contextmanager
    def edit_data(self) -> List['DXFEntity']:
        """ Context manager which yields all the group entities as
        standard Python list::

            with group.edit_data() as data:
               # add new entities to a group
               data.append(modelspace.add_line((0, 0), (3, 0)))
               # remove last entity from a group
               data.pop()

        """
        data = list(self)
        yield data
        self.set_data(data)

    def set_data(self, entities: Iterable['DXFEntity']) -> None:
        """  Set `entities` as new group content, entities should be an iterable
        :class:`DXFGraphic` or inherited (LINE, CIRCLE, ...).
        Raises :class:`DXFValueError` if not all entities be on the same layout
        (modelspace or any paperspace layout but not block)

        """
        entities = list(entities)
        if not all_entities_on_same_layout(entities):
            raise DXFValueError(
                "All entities have to be on the same layout (modelspace or any "
                "paperspace layout but not a block)."
            )
        self.clear()
        self._data = entities

    def extend(self, entities: Iterable['DXFEntity']) -> None:
        """ Add `entities` to :class:`DXFGroup`. """
        self._data.extend(entities)

    def clear(self) -> None:
        """ Remove all entities from :class:`DXFGroup`, does not delete any
        drawing entities referenced by this group.

        """
        self._data = []

    def audit(self, auditor: 'Auditor') -> None:
        """ Remove invalid handles from :class:`DXFGroup`.

        Invalid handles are: deleted entities, not all entities in the same
        layout or entities in a block layout.

        """
        # Remove destroyed or invalid entities:
        db = auditor.entitydb
        self._data = list(self._filter_invalid_entities(db))
        if not all_entities_on_same_layout(self._data):
            auditor.fixed_error(
                code=AuditError.GROUP_ENTITIES_IN_DIFFERENT_LAYOUTS,
                message=f'Cleared {str(self)}, not all entities are located in '
                        f'the same layout.',
            )
            self.clear()

    def _has_valid_owner(self, entity, db: 'EntityDB') -> bool:
        # no owner -> no layout association
        if entity.dxf.owner is None:
            return False
        owner = db.get(entity.dxf.owner)
        # owner does not exist or is destroyed -> no layout association
        if owner is None or not owner.is_alive:
            return False
        # owner block_record.layout is 0 if entity is in a block definition,
        # which is not allowed:
        valid = owner.dxf.layout != '0'
        if not valid:
            logger.debug(
                f'Removed {str(entity)} from {str(self)}, because entity is '
                f'located in a block layout.')
        return valid

    def _filter_invalid_entities(self, db: 'EntityDB') -> Iterable['DXFEntity']:
        assert db is not None
        for e in self._data:
            if e is None:
                continue
            if e.is_alive and self._has_valid_owner(e, db):
                yield e


def all_entities_on_same_layout(entities: Iterable['DXFEntity']):
    """ Check if all entities are on the same layout (model space or any paper
    layout but not block).

    """
    owners = set(entity.dxf.owner for entity in entities)
    # 0 for no entities; 1 for all entities on the same layout
    return len(owners) < 2


class GroupCollection(ObjectCollection):
    def __init__(self, doc: 'Drawing'):
        super().__init__(doc, dict_name='ACAD_GROUP', object_type='GROUP')
        self._next_unnamed_number = 0

    def groups(self) -> Iterable[DXFGroup]:
        """ Iterable of all existing groups. """
        for name, group in self:
            yield group

    def next_name(self) -> str:
        name = self._next_name()
        while name in self:
            name = self._next_name()
        return name

    def _next_name(self) -> str:
        self._next_unnamed_number += 1
        return f"*A{self._next_unnamed_number}"

    def new(self, name: str = None, description: str = "",
            selectable: bool = True) -> DXFGroup:
        r""" Creates a new group. If `name` is ``None`` an unnamed group is
        created, which has an automatically generated name like "\*Annnn".

        Args:
            name: group name as string
            description: group description as string
            selectable: group is selectable if ``True``

        """
        if name in self:
            raise DXFValueError(f"GROUP '{name}' already exists.")

        if name is None:
            name = self.next_name()
            unnamed = 1
        else:
            unnamed = 0
        # The group name isn't stored in the group entity itself.
        dxfattribs = {
            'description': description,
            'unnamed': unnamed,
            'selectable': int(bool(selectable)),
        }
        return cast(DXFGroup, self._new(name, dxfattribs))

    def delete(self, group: Union[DXFGroup, str]) -> None:
        """ Delete `group`, `group` can be an object of type :class:`DXFGroup`
        or a group name as string.

        """
        # Delete group by name:
        if isinstance(group, str):
            name = group
        elif group.dxftype() == 'GROUP':
            name = get_group_name(group, self.entitydb)
        else:
            raise TypeError(group.dxftype())

        if name in self:
            super().delete(name)
        else:
            raise DXFValueError("GROUP not in group table registered.")

    def audit(self, auditor: 'Auditor') -> None:
        """ Removes empty groups and invalid handles from all groups. """
        trash = []
        for name, group in self:
            group.audit(auditor)
            if not len(group):  # remove empty group
                # do not delete groups while iterating over groups!
                trash.append(name)

        # now delete empty groups
        for name in trash:
            auditor.fixed_error(
                code=AuditError.REMOVE_EMPTY_GROUP,
                message=f'Removed empty group "{name}".',
            )
            self.delete(name)


def get_group_name(group: DXFGroup, db: 'EntityDB') -> str:
    """ Get name of `group`. """
    group_table = cast('Dictionary', db[group.dxf.owner])
    for name, entity in group_table.items():
        if entity is group:
            return name



# Created: 06.2020
# Copyright (c) 2020, Matthew Broadway
# License: MIT License
import enum
import re
from math import radians
from typing import Union, Tuple, Dict, Iterable, List, Optional, Callable

from ezdxf.entities.text import plain_text

import ezdxf.lldxf.const as DXFConstants
from ezdxf.addons.drawing.backend import Backend
from ezdxf.addons.drawing.debug_utils import draw_rect
from ezdxf.addons.drawing import fonts
from ezdxf.entities import MText, Text, Attrib, AttDef
from ezdxf.math import Matrix44, Vec3, sign

"""
Search google for 'typography' or 'font anatomy' for explanations of terms like 'baseline' and 'x-height'

A Visual Guide to the Anatomy of Typography: https://visme.co/blog/type-anatomy/
Anatomy of a Character: https://www.fonts.com/content/learning/fontology/level-1/type-anatomy/anatomy
"""


@enum.unique
class HAlignment(enum.Enum):
    LEFT = 0
    CENTER = 1
    RIGHT = 2


@enum.unique
class VAlignment(enum.Enum):
    TOP = 0  # the top of capital letters or letters with ascenders (like 'b')
    LOWER_CASE_CENTER = 1  # the midpoint between the baseline and the x-height
    BASELINE = 2  # the line which text rests on, characters with descenders (like 'p') are partially below this line
    BOTTOM = 3  # the lowest point on a character with a descender (like 'p')


Alignment = Tuple[HAlignment, VAlignment]
AnyText = Union[Text, MText, Attrib, AttDef]

# multiple of cap_height between the baseline of the previous line and the baseline of the next line
DEFAULT_LINE_SPACING = 5 / 3

DXF_MTEXT_ALIGNMENT_TO_ALIGNMENT: Dict[int, Alignment] = {
    DXFConstants.MTEXT_TOP_LEFT: (HAlignment.LEFT, VAlignment.TOP),
    DXFConstants.MTEXT_TOP_CENTER: (HAlignment.CENTER, VAlignment.TOP),
    DXFConstants.MTEXT_TOP_RIGHT: (HAlignment.RIGHT, VAlignment.TOP),
    DXFConstants.MTEXT_MIDDLE_LEFT: (HAlignment.LEFT, VAlignment.LOWER_CASE_CENTER),
    DXFConstants.MTEXT_MIDDLE_CENTER: (HAlignment.CENTER, VAlignment.LOWER_CASE_CENTER),
    DXFConstants.MTEXT_MIDDLE_RIGHT: (HAlignment.RIGHT, VAlignment.LOWER_CASE_CENTER),
    DXFConstants.MTEXT_BOTTOM_LEFT: (HAlignment.LEFT, VAlignment.BOTTOM),
    DXFConstants.MTEXT_BOTTOM_CENTER: (HAlignment.CENTER, VAlignment.BOTTOM),
    DXFConstants.MTEXT_BOTTOM_RIGHT: (HAlignment.RIGHT, VAlignment.BOTTOM)
}
assert len(DXF_MTEXT_ALIGNMENT_TO_ALIGNMENT) == len(DXFConstants.MTEXT_ALIGN_FLAGS)


class FontMeasurements:
    def __init__(self, baseline: float, cap_height: float, x_height: float, descender_height: float):
        self.baseline = baseline
        self.cap_height = cap_height
        self.x_height = x_height
        self.descender_height = descender_height

    def __eq__(self, other):
        return (isinstance(other, FontMeasurements) and
                self.baseline == other.baseline and
                self.cap_height == other.cap_height and
                self.x_height == other.x_height and
                self.descender_height == other.descender_height)

    def scale_from_baseline(self, desired_cap_height: float) -> "FontMeasurements":
        scale = desired_cap_height / self.cap_height
        return FontMeasurements(
            baseline=self.baseline,
            cap_height=desired_cap_height,
            x_height=self.x_height * scale,
            descender_height=self.descender_height * scale,
        )

    @property
    def cap_top(self) -> float:
        return self.baseline + self.cap_height

    @property
    def x_top(self) -> float:
        return self.baseline + self.x_height

    @property
    def bottom(self) -> float:
        return self.baseline - self.descender_height


def _get_rotation(text: AnyText) -> Matrix44:
    if isinstance(text, Text):
        return Matrix44.axis_rotate(text.dxf.extrusion, radians(text.dxf.rotation))
    if isinstance(text, MText):
        return Matrix44.axis_rotate(Vec3(0, 0, 1), radians(text.get_rotation()))
    elif isinstance(text, (Attrib, AttDef)):
        return Matrix44()
    else:
        raise TypeError(type(text))


def _get_alignment(text: AnyText) -> Alignment:
    if isinstance(text, Text):
        # for the purposes of rendering, the anchor is always the bottom left, text.get_align() should be ignored
        return HAlignment.LEFT, VAlignment.BASELINE
    elif isinstance(text, MText):
        return DXF_MTEXT_ALIGNMENT_TO_ALIGNMENT[text.dxf.attachment_point]
    elif isinstance(text, (Attrib, AttDef)):
        return HAlignment.LEFT, VAlignment.BASELINE
    else:
        raise TypeError(type(text))


def _get_cap_height(text: AnyText) -> float:
    if isinstance(text, (Text, Attrib, AttDef)):
        return text.dxf.height
    elif isinstance(text, MText):
        return text.dxf.char_height
    else:
        raise TypeError(type(text))


def _get_line_spacing(text: AnyText, cap_height: float) -> float:
    if isinstance(text, (Attrib, AttDef, Text)):
        return 0.0
    elif isinstance(text, MText):
        return cap_height * DEFAULT_LINE_SPACING * text.dxf.line_spacing_factor
    else:
        raise TypeError(type(text))


def _split_multiline_text(text: str, box_width: Optional[float], get_text_width: Callable[[str], float]) -> List[str]:
    """
    This isn't the most straightforward word wrapping algorithm, but it aims to match the behavior of AutoCAD
    """
    if not text or text.isspace():
        return []
    manual_lines = re.split(r'(\n)', text)  # includes \n as it's own token
    tokens = [t for line in manual_lines for t in re.split(r'(\s+)', line) if t]
    lines = []
    current_line = ''
    line_just_wrapped = False

    for t in tokens:
        on_first_line = not lines
        if t == '\n' and line_just_wrapped:
            continue
        line_just_wrapped = False
        if t == '\n':
            lines.append(current_line.rstrip())
            current_line = ''
        elif t.isspace():
            if current_line or on_first_line:
                current_line += t
        else:
            if box_width is not None and get_text_width(current_line + t) > box_width:
                if not current_line:
                    current_line += t
                else:
                    lines.append(current_line.rstrip())
                    current_line = t
                    line_just_wrapped = True
            else:
                current_line += t

    if current_line and not current_line.isspace():
        lines.append(current_line.rstrip())
    return lines


def _split_into_lines(entity: AnyText, box_width: Optional[float], get_text_width: Callable[[str], float]) -> List[str]:
    if isinstance(entity, AttDef):
        # ATTDEF outside of an Insert renders the tag rather than the value
        text = plain_text(entity.dxf.tag)
    else:
        text = entity.plain_text()
    if isinstance(entity, (Text, Attrib, AttDef)):
        assert '\n' not in text
        return [text]
    else:
        return _split_multiline_text(text, box_width, get_text_width)


def _get_text_width(text: AnyText) -> Optional[float]:
    if isinstance(text, (Attrib, AttDef, Text)):
        return None
    elif isinstance(text, MText):
        width = text.dxf.width
        return None if width == 0.0 else width
    else:
        raise TypeError(type(text))


def _get_extra_transform(text: AnyText) -> Matrix44:
    extra_transform = Matrix44()
    if isinstance(text, Text):
        # ALIGNED: scaled to fit in the text box (aspect ratio preserved). Does not have to be handled specially.
        # FIT: scaled to fit in the text box (aspect ratio *not* preserved). Handled by dxf.width
        scale_x = text.dxf.width  # 'width' is the width *scale factor* so 1.0 by default
        scale_y = 1
        if text.dxf.text_generation_flag & DXFConstants.MIRROR_X:
            scale_x *= -1
        if text.dxf.text_generation_flag & DXFConstants.MIRROR_Y:
            scale_y *= -1

        # magnitude of extrusion does not have any effect. An extrusion of (0, 0, 0) acts like (0, 0, 1)
        scale_x *= sign(text.dxf.extrusion.z)

        if scale_x != 1 or scale_y != 1:
            extra_transform = Matrix44.scale(scale_x, scale_y)

    elif isinstance(text, MText):
        # not sure about the rationale behind this but it does match AutoCAD behavior...
        scale_y = sign(text.dxf.extrusion.z)
        if scale_y != 1:
            extra_transform = Matrix44.scale(1, scale_y)

    return extra_transform


def _apply_alignment(alignment: Alignment,
                     line_widths: List[float],
                     line_spacing: float,
                     box_width: Optional[float],
                     font_measurements: FontMeasurements) -> Tuple[Tuple[float, float], List[float], List[float]]:
    if not line_widths:
        return (0, 0), [], []

    halign, valign = alignment
    line_ys = [-font_measurements.baseline -
               (font_measurements.cap_height + i * line_spacing)
               for i in range(len(line_widths))]

    if box_width is None:
        box_width = max(line_widths)

    last_baseline = line_ys[-1]

    if halign == HAlignment.LEFT:
        anchor_x = 0
        line_xs = [0] * len(line_widths)
    elif halign == HAlignment.CENTER:
        anchor_x = box_width / 2
        line_xs = [anchor_x - w / 2 for w in line_widths]
    elif halign == HAlignment.RIGHT:
        anchor_x = box_width
        line_xs = [anchor_x - w for w in line_widths]
    else:
        raise ValueError(halign)

    if valign == VAlignment.TOP:
        anchor_y = 0
    elif valign == VAlignment.LOWER_CASE_CENTER:
        first_line_lower_case_top = line_ys[0] + font_measurements.x_height
        anchor_y = (first_line_lower_case_top + last_baseline) / 2
    elif valign == VAlignment.BASELINE:
        anchor_y = last_baseline
    elif valign == VAlignment.BOTTOM:
        anchor_y = last_baseline - font_measurements.descender_height
    else:
        raise ValueError(valign)

    return (anchor_x, anchor_y), line_xs, line_ys


def _get_wcs_insert(text: AnyText) -> Vec3:
    if isinstance(text, Text):
        return text.ocs().to_wcs(text.dxf.insert)
    else:
        return text.dxf.insert


def simplified_text_chunks(text: AnyText, out: Backend,
                           *,
                           font: fonts.Font = None,
                           debug_draw_rect: bool = False) -> Iterable[Tuple[str, Matrix44, float]]:
    """
    Splits a complex text entity into simple chunks of text which can all be rendered the same way:
    render the string (which will not contain any newlines) with the given cap_height with (left, baseline) at (0, 0)
    then transform it with the given matrix to move it into place.
    """
    alignment = _get_alignment(text)
    box_width = _get_text_width(text)

    cap_height = _get_cap_height(text)
    lines = _split_into_lines(text, box_width, lambda s: out.get_text_line_width(s, cap_height, font=font))
    line_spacing = _get_line_spacing(text, cap_height)
    line_widths = [out.get_text_line_width(line, cap_height, font=font) for line in lines]
    font_measurements = out.get_font_measurements(cap_height, font=font)
    anchor, line_xs, line_ys = \
        _apply_alignment(alignment, line_widths, line_spacing, box_width, font_measurements)
    rotation = _get_rotation(text)
    extra_transform = _get_extra_transform(text)
    insert = _get_wcs_insert(text)

    whole_text_transform = (
            Matrix44.translate(-anchor[0], -anchor[1], 0) @
            extra_transform @
            rotation @
            Matrix44.translate(*insert.xyz)
    )
    for i, (line, line_x, line_y) in enumerate(zip(lines, line_xs, line_ys)):
        transform = Matrix44.translate(line_x, line_y, 0) @ whole_text_transform
        yield line, transform, cap_height

        if debug_draw_rect:
            width = out.get_text_line_width(line, cap_height, font)
            ps = list(transform.transform_vertices([Vec3(0, 0, 0), Vec3(width, 0, 0), Vec3(width, cap_height, 0),
                                                    Vec3(0, cap_height, 0), Vec3(0, 0, 0)]))
            draw_rect(ps, '#ff0000', out)



# Copyright (c) 2020, Matthew Broadway
# License: MIT License
import math
from typing import Iterable, TYPE_CHECKING, Optional, Dict, Sequence
import warnings
from collections import defaultdict
from functools import lru_cache

import matplotlib.pyplot as plt
from matplotlib.collections import LineCollection
from matplotlib.font_manager import FontProperties
from matplotlib.lines import Line2D
from matplotlib.patches import Circle, PathPatch
from matplotlib.path import Path
from matplotlib.textpath import TextPath
import numpy as np

from ezdxf.addons.drawing.backend import Backend, prepare_string_for_rendering
from ezdxf.addons.drawing.properties import Properties
from ezdxf.addons.drawing.text import FontMeasurements
from ezdxf.addons.drawing.type_hints import Color
from ezdxf.addons.drawing import fonts
from ezdxf.math import Vec3, Matrix44
from ezdxf.render import Command
from ezdxf.render.linetypes import LineTypeRenderer as EzdxfLineTypeRenderer
from .matplotlib_hatch import HATCH_NAME_MAPPING
from .line_renderer import AbstractLineRenderer

if TYPE_CHECKING:
    from ezdxf.eztypes import Layout

# matplotlib docs: https://matplotlib.org/index.html

# line style:
# https://matplotlib.org/api/_as_gen/matplotlib.lines.Line2D.html#matplotlib.lines.Line2D.set_linestyle
# https://matplotlib.org/gallery/lines_bars_and_markers/linestyles.html

# line width:
# https://matplotlib.org/api/_as_gen/matplotlib.lines.Line2D.html#matplotlib.lines.Line2D.set_linewidth
# points unit (pt), 1pt = 1/72 inch, 1pt = 0.3527mm
POINTS = 1.0 / 0.3527  # mm -> points
CURVE4x3 = (Path.CURVE4, Path.CURVE4, Path.CURVE4)
MATPLOTLIB_DEFAULT_PARAMS = {}


def get_params(params: Optional[Dict]) -> Dict:
    default_params = dict(MATPLOTLIB_DEFAULT_PARAMS)
    default_params.update(params or {})
    return default_params


class MatplotlibBackend(Backend):
    def __init__(self, ax: plt.Axes,
                 *,
                 adjust_figure: bool = True,
                 font: FontProperties = FontProperties(),
                 use_text_cache: bool = True,
                 params: Dict = None,
                 ):
        super().__init__(get_params(params))
        self.ax = ax
        self._adjust_figure = adjust_figure
        self._scale_dashes_backup = plt.rcParams['lines.scale_dashes']
        # Disable internal line style scaling by matplotlib
        plt.rcParams['lines.scale_dashes'] = False

        # like set_axis_off, except that the face_color can still be set
        self.ax.xaxis.set_visible(False)
        self.ax.yaxis.set_visible(False)
        for s in self.ax.spines.values():
            s.set_visible(False)

        self.ax.autoscale(False)
        self.ax.set_aspect('equal', 'datalim')
        self._current_z = 0
        self._text_renderer = TextRenderer(font, use_text_cache)

        # Setup line rendering component:
        if self.linetype_renderer == "ezdxf":
            # This linetype renderer should only be used by "hardcopy" backends!
            # It is just too slow for interactive backends, and the result of
            # the matplotlib line rendering is optimized for displays.
            self._line_renderer = EzdxfLineRenderer(self)
        else:
            self._line_renderer = InternalLineRenderer(self)

    def clear_text_cache(self):
        self._text_renderer.clear_cache()

    def _get_z(self) -> int:
        z = self._current_z
        self._current_z += 1
        return z

    def set_background(self, color: Color):
        self.ax.set_facecolor(color)

    def draw_point(self, pos: Vec3, properties: Properties):
        """ Draw a real dimensionless point. """
        color = properties.color
        self.ax.scatter([pos.x], [pos.y], s=0.1, c=color, zorder=self._get_z())

    def draw_line(self, start: Vec3, end: Vec3, properties: Properties):
        # matplotlib draws nothing for a zero-length line:
        if start.isclose(end):
            self.draw_point(start, properties)
        else:
            self._line_renderer.draw_line(start, end, properties, self._get_z())

    def draw_path(self, path, properties: Properties):
        self._line_renderer.draw_path(path, properties, self._get_z())

    def draw_filled_paths(self, paths: Sequence,
                          holes: Sequence, properties: Properties):
        fill, hatch = self._get_filling(properties)
        if fill is False and hatch is None:
            return

        vertices = []
        codes = []
        for path in paths:
            v1, c1 = _get_path_patch_data(path.counter_clockwise())
            vertices.extend(v1)
            codes.extend(c1)

        for hole in holes:
            v1, c1 = _get_path_patch_data(hole.clockwise())
            vertices.extend(v1)
            codes.extend(c1)

        patch = PathPatch(
            Path(vertices, codes),
            color=properties.color,
            fill=fill,
            hatch=hatch,
            zorder=self._get_z()
        )
        self.ax.add_patch(patch)

    def draw_filled_polygon(self, points: Iterable[Vec3],
                            properties: Properties):
        self.ax.fill(*zip(*((p.x, p.y) for p in points)),
                     color=properties.color, zorder=self._get_z())

    def draw_text(self, text: str, transform: Matrix44, properties: Properties,
                  cap_height: float):
        if not text.strip():
            return  # no point rendering empty strings
        font_properties = self.get_font_properties(properties.font)
        text = prepare_string_for_rendering(text, self.current_entity.dxftype())
        transformed_path = _transform_path(
            self._text_renderer.get_text_path(text, font_properties),
            Matrix44.scale(
                self._text_renderer.get_scale(cap_height,
                                              font_properties)) @ transform
        )
        self.ax.add_patch(
            PathPatch(transformed_path, facecolor=properties.color, linewidth=0,
                      zorder=self._get_z()))

    @lru_cache(maxsize=256)  # fonts.Font is a named tuple
    def get_font_properties(self, font: fonts.Font) -> FontProperties:
        font_properties = self._text_renderer.default_font
        if font:
            # Font-definitions are created by the matplotlib FontManger(),
            # but stored as json file and could be altered by an user:
            try:
                font_properties = FontProperties(
                    family=font.family,
                    style=font.style,
                    stretch=font.stretch,
                    weight=font.weight,
                )
            except ValueError:
                pass
        return font_properties

    def get_font_measurements(self, cap_height: float,
                              font: str = None) -> FontMeasurements:
        return self._text_renderer.get_font_measurements(
            self.get_font_properties(font)).scale_from_baseline(
            desired_cap_height=cap_height)

    def get_text_line_width(self, text: str, cap_height: float,
                            font: str = None) -> float:
        if not text.strip():
            return 0
        dxftype = self.current_entity.dxftype() if self.current_entity else 'TEXT'
        text = prepare_string_for_rendering(text, dxftype)
        font_properties = self.get_font_properties(font)
        path = self._text_renderer.get_text_path(text, font_properties)
        return max(x for x, y in path.vertices) * self._text_renderer.get_scale(
            cap_height, font_properties)

    def clear(self):
        self.ax.clear()

    def finalize(self):
        super().finalize()
        self.ax.autoscale(True)
        if self._adjust_figure:
            minx, maxx = self.ax.get_xlim()
            miny, maxy = self.ax.get_ylim()
            data_width, data_height = maxx - minx, maxy - miny
            if not math.isclose(data_width, 0):
                width, height = plt.figaspect(data_height / data_width)
                self.ax.get_figure().set_size_inches(width, height,
                                                     forward=True)
        plt.rcParams['lines.scale_dashes'] = self._scale_dashes_backup

    def _get_filling(self, properties: Properties):
        fill = True
        hatch = None
        name = properties.filling.name.upper()
        if properties.filling.type == 1 and name != 'SOLID':
            if self.hatch_pattern == 0:
                # Disable hatch patterns
                fill = False
                hatch = False
            elif self.hatch_pattern == 1:
                # Use predefined hatch pattern by name matching:
                fill = False
                hatch = HATCH_NAME_MAPPING.get(name, r'\\\\')
            else:
                # Draw hatch pattern as solid filling
                fill = True
                hatch = False
        return fill, hatch


def _transform_path(path: Path, transform: Matrix44) -> Path:
    vertices = transform.transform_vertices(
        [Vec3(x, y) for x, y in path.vertices])
    return Path([(v.x, v.y) for v in vertices], path.codes)


class TextRenderer:
    def __init__(self, font: FontProperties, use_cache: bool):
        self._default_font = font
        self._use_cache = use_cache

        # Each font has its own text path cache
        # key is hash(FontProperties)
        self._text_path_cache: Dict[
            int, Dict[str, TextPath]] = defaultdict(dict)

        # Each font has its own font measurements cache
        # key is hash(FontProperties)
        self._font_measurement_cache: Dict[
            int, FontMeasurements] = {}

    @property
    def default_font(self) -> FontProperties:
        return self._default_font

    def clear_cache(self):
        self._text_path_cache.clear()

    def get_scale(self, desired_cap_height: float,
                  font: FontProperties) -> float:
        return desired_cap_height / self.get_font_measurements(font).cap_height

    def get_font_measurements(self, font: FontProperties) -> FontMeasurements:
        # None is the default font.
        key = hash(font)
        measurements = self._font_measurement_cache.get(key)
        if measurements is None:
            upper_x = self.get_text_path('X', font).vertices[:, 1].tolist()
            lower_x = self.get_text_path('x', font).vertices[:, 1].tolist()
            lower_p = self.get_text_path('p', font).vertices[:, 1].tolist()
            baseline = min(lower_x)
            measurements = FontMeasurements(
                baseline=baseline,
                cap_height=max(upper_x) - baseline,
                x_height=max(lower_x) - baseline,
                descender_height=baseline - min(lower_p)
            )
            self._font_measurement_cache[key] = measurements
        return measurements

    def get_text_path(self, text: str, font: FontProperties) -> TextPath:
        # None is the default font
        cache = self._text_path_cache[hash(font)]  # defaultdict(dict)
        path = cache.get(text, None)
        if path is None:
            if font is None:
                font = self._default_font
            # must replace $ with \$ to avoid matplotlib interpreting it as math text
            path = TextPath((0, 0), text.replace('$', '\\$'), size=1,
                            prop=font, usetex=False)
            if self._use_cache:
                cache[text] = path
        return path


def _get_path_patch_data(path):
    codes = [Path.MOVETO]
    vertices = [path.start]
    for cmd in path:
        if cmd.type == Command.LINE_TO:
            codes.append(Path.LINETO)
            vertices.append(cmd.end)
        elif cmd.type == Command.CURVE_TO:
            codes.extend(CURVE4x3)
            vertices.extend((cmd.ctrl1, cmd.ctrl2, cmd.end))
        else:
            raise ValueError(f'Invalid command: {cmd.type}')
    return [(p.x, p.y) for p in vertices], codes


def qsave(layout: 'Layout', filename: str, *,
          bg: Optional[Color] = None,
          fg: Optional[Color] = None,
          dpi: int = 300,
          backend: str = 'agg',
          ltype=None,  # deprecated
          lineweight_scaling=None,  # deprecated
          params: dict = None,
          ) -> None:
    """ Quick and simplified render export by matplotlib.

    Args:
        layout: modelspace or paperspace layout to export
        filename: export filename, file extension determines the format e.g.
            "image.png" to save in PNG format.
        bg: override default background color in hex format #RRGGBB or #RRGGBBAA,
            e.g. use bg="#FFFFFF00" to get a transparent background and a black
            foreground color (ACI=7), because a white background #FFFFFF gets a
            black foreground color or vice versa bg="#00000000" for a transparent
            (black) background and a white foreground color.
        fg: override default foreground color in hex format #RRGGBB or #RRGGBBAA,
            requires also `bg` argument. There is no explicit foreground color
            in DXF defined (also not a background color), but the ACI color 7
            has already a variable color value, black on a light background and
            white on a dark background, this argument overrides this (ACI=7)
            default color value.
        dpi: image resolution (dots per inches).
        backend: the matplotlib rendering backend to use (agg, cairo, svg etc)
            (see documentation for `matplotlib.use() <https://matplotlib.org/3.1.1/api/matplotlib_configuration_api.html?highlight=matplotlib%20use#matplotlib.use>`_
            for a complete list of backends)
        ltype: deprecated, use :code:`params={"linetype_renderer": "ezdxf"}`
        lineweight_scaling: deprecated, use :code:`params={"lineweight_scaling": 0}`
        params: matplotlib backend parameters

    .. versionadded:: 0.14

    .. versionchanged:: 0.15

        deprecated arguments `ltype` and `lineweight_scaling` will be removed in
        v0.16, added argument `params` to pass parameters to the matplotlib
        backend.

    """
    from .properties import RenderContext
    from .frontend import Frontend
    import matplotlib

    # Set the backend to prevent warnings about GUIs being opened from a thread
    # other than the main thread.
    old_backend = matplotlib.get_backend()
    matplotlib.use(backend)
    params = params or {}

    # Let the user choose a minimum lineweight:
    if 'min_lineweight' not in params:
        # If not set by user, use ~1 pixel
        params['min_lineweight'] = 72 / dpi

    if ltype is not None:
        params['linetype_renderer'] = ltype
        warnings.warn(
            'The "ltype" argument is deprecated use the "params" dict '
            'to pass arguments to the MatplotlibBackend, '
            'will be removed in v0.16.', DeprecationWarning)
    if lineweight_scaling is not None:
        params['lineweight_scaling'] = lineweight_scaling
        warnings.warn(
            'The "lineweight_scaling" argument is deprecated use the '
            '"params" dict to pass arguments to the MatplotlibBackend, '
            'will be removed in v0.16.', DeprecationWarning)

    try:
        fig: plt.Figure = plt.figure()
        ax: plt.Axes = fig.add_axes((0, 0, 1, 1))
        ctx = RenderContext(layout.doc)
        ctx.set_current_layout(layout)
        if bg is not None:
            ctx.current_layout.set_colors(bg, fg)
        out = MatplotlibBackend(ax, params=params)
        Frontend(ctx, out).draw_layout(layout, finalize=True)
        # transparent=True sets the axes color to fully transparent
        # facecolor sets the figure color
        # (semi-)transparent axes colors do not produce transparent outputs
        # but (semi-)transparent figure colors do.
        fig.savefig(filename, dpi=dpi,
                    facecolor=ax.get_facecolor(), transparent=True)
        plt.close(fig)
    finally:
        matplotlib.use(old_backend)


class MatplotlibLineRenderer(AbstractLineRenderer):
    @property
    def lineweight_scaling(self) -> float:
        return self._backend.lineweight_scaling * POINTS

    # noinspection PyUnresolvedReferences
    @property
    def ax(self) -> plt.Axes:
        return self._backend.ax  # MatplotlibBackend


# Scaling factor for internal renderer, just guessing here:
ISO_LIN_PATTERN_FACTOR = 3.0 * POINTS
ANSI_LIN_PATTERN_FACTOR = ISO_LIN_PATTERN_FACTOR * 2.54


class InternalLineRenderer(MatplotlibLineRenderer):
    """ matplotlib internal linetype rendering, which is oriented on the output
    medium and dpi: This method is simpler and faster but may not replicate the
    results of CAD applications.
    """

    def draw_line(self, start: Vec3,
                  end: Vec3, properties: Properties, z: float):
        self.ax.add_line(
            Line2D(
                (start.x, end.x), (start.y, end.y),
                linewidth=self.lineweight(properties),
                linestyle=self.linetype(properties),
                color=properties.color,
                zorder=z,
            ))

    def draw_path(self, path, properties: Properties, z: float):
        vertices, codes = _get_path_patch_data(path)
        patch = PathPatch(
            Path(vertices, codes),
            linewidth=self.lineweight(properties),
            linestyle=self.linetype(properties),
            fill=False,
            color=properties.color,
            zorder=z
        )
        self.ax.add_patch(patch)

    @property
    def measurement_scale(self) -> float:
        return ISO_LIN_PATTERN_FACTOR if self.measurement \
            else ANSI_LIN_PATTERN_FACTOR

    def create_pattern(self, properties: Properties, scale: float):
        """ Return matplotlib line style tuple: (offset, on_off_sequence) or
        "solid".

        See examples: https://matplotlib.org/gallery/lines_bars_and_markers/linestyles.html

        """
        if len(properties.linetype_pattern) < 2:
            return 'solid'
        else:
            pattern = np.round(np.array(properties.linetype_pattern) * scale)
            pattern = [max(element, 1) for element in pattern]
            if len(pattern) % 2:
                pattern.pop()
            return 0, pattern


class EzdxfLineRenderer(MatplotlibLineRenderer):
    """ Replicate AutoCAD linetype rendering oriented on drawing units and
    various ltscale factors. This rendering method break lines into small
    segments which causes a longer rendering time!
    """

    def draw_line(self, start: Vec3, end: Vec3,
                  properties: Properties, z: int):
        pattern = self.pattern(properties)
        lineweight = self.lineweight(properties)
        render_linetypes = bool(self.linetype_scaling)
        color = properties.color
        if len(pattern) < 2 or not render_linetypes:
            self.ax.add_line(
                Line2D(
                    (start.x, end.x), (start.y, end.y),
                    linewidth=lineweight,
                    color=color,
                    zorder=z,
                ))
        else:
            renderer = EzdxfLineTypeRenderer(pattern)
            lines = LineCollection(
                [((s.x, s.y), (e.x, e.y))
                 for s, e in renderer.line_segment(start, end)],
                linewidths=lineweight, color=color, zorder=z
            )
            lines.set_capstyle('butt')
            self.ax.add_collection(lines)

    def draw_path(self, path, properties: Properties, z: int):
        pattern = self.pattern(properties)
        lineweight = self.lineweight(properties)
        render_linetypes = bool(self.linetype_scaling)
        color = properties.color
        if len(pattern) < 2 or not render_linetypes:
            vertices, codes = _get_path_patch_data(path)
            patch = PathPatch(
                Path(vertices, codes),
                linewidth=lineweight,
                color=color,
                fill=False,
                zorder=z
            )
            self.ax.add_patch(patch)
        else:
            renderer = EzdxfLineTypeRenderer(pattern)
            segments = renderer.line_segments(path.flattening(
                self.max_flattening_distance, segments=16))
            lines = LineCollection(
                [((s.x, s.y), (e.x, e.y)) for s, e in segments],
                linewidths=lineweight, color=color, zorder=z
            )
            lines.set_capstyle('butt')
            self.ax.add_collection(lines)



# Purpose: julian date
# Created: 21.03.2011
# Copyright (c) 2011-2018, Manfred Moitzi
# License: MIT License
from typing import Tuple
from math import floor
from datetime import datetime


def frac(number: float) -> float:
    return number - floor(number)


class JulianDate:
    def __init__(self, date: datetime):
        self.date = date
        self.result = self.julian_date() + self.fractional_day()  # type: float

    def fractional_day(self) -> float:
        seconds = self.date.hour * 3600. + self.date.minute * 60. + self.date.second
        return seconds / 86400.

    def julian_date(self) -> float:
        y = self.date.year + (float(self.date.month) - 2.85) / 12.
        A = floor(367. * y) - 1.75 * floor(y) + self.date.day
        B = floor(A) - 0.75 * floor(y / 100.)
        return floor(B) + 1721115.


class CalendarDate:
    def __init__(self, juliandate: float):
        self.jdate = juliandate
        year, month, day = self.get_date()
        hour, minute, second = frac2time(self.jdate)
        self.result = datetime(year, month, day, hour, minute, second)

    def get_date(self) -> Tuple[int, int, int]:
        Z = floor(self.jdate)

        if Z < 2299161:
            A = Z  # julian calendar
        else:
            g = floor((Z - 1867216.25) / 36524.25)  # gregorian calendar
            A = Z + 1. + g - floor(g / 4.)

        B = A + 1524.
        C = floor((B - 122.1) / 365.25)
        D = floor(365.25 * C)
        E = floor((B - D) / 30.6001)

        day = B - D - floor(30.6001 * E)
        month = E - 1 if E < 14 else E - 13
        year = C - 4716 if month > 2 else C - 4715
        return int(year), int(month), int(day)


def frac2time(jdate) -> Tuple[int, int, int]:
    seconds = int(frac(jdate) * 86400.)
    hour = int(seconds / 3600)
    seconds = seconds % 3600
    minute = int(seconds / 60)
    second = seconds % 60
    return hour, minute, second


def juliandate(date: datetime) -> float:
    return JulianDate(date).result


def calendardate(juliandate: float) -> datetime:
    return CalendarDate(juliandate).result



# Copyright (c) 2010-2020, Manfred Moitzi
# License: MIT License
from typing import TYPE_CHECKING, Optional
import math
from ezdxf.math import Vec2, intersection_line_line_2d
from .construct2d import is_point_left_of_line, TOLERANCE
from .bbox import BoundingBox2d

if TYPE_CHECKING:
    from ezdxf.eztypes import Vertex


class ParallelRaysError(ArithmeticError):
    pass


HALF_PI = math.pi / 2.
THREE_PI_HALF = 1.5 * math.pi
DOUBLE_PI = math.pi * 2.


class ConstructionRay:
    """
    Infinite 2D construction ray as immutable object.

    Args:
        p1: definition point 1
        p2: ray direction as 2nd point or ``None``
        angle: ray direction as angle in radians or ``None``

    """

    def __init__(self, p1: 'Vertex', p2: 'Vertex' = None, angle: float = None):
        self._location = Vec2(p1)
        if p2 is not None:
            p2 = Vec2(p2)
            if self._location.x < p2.x:
                self._direction = (p2 - self._location).normalize()
            else:
                self._direction = (self._location - p2).normalize()
            self._angle = self._direction.angle
        elif angle is not None:
            self._angle = angle
            self._direction = Vec2.from_angle(angle)
        else:
            raise ValueError('p2 or angle required.')

        if math.isclose(self._direction.x, 0., abs_tol=1e-12):
            self._slope = None
            self._yof0 = None
        else:
            self._slope = self._direction.y / self._direction.x
            self._yof0 = self._location.y - self._slope * self._location.x
        self._is_vertical = self._slope is None
        self._is_horizontal = math.isclose(self._direction.y, 0., abs_tol=1e-12)

    @property
    def location(self) -> Vec2:
        """ Location vector as :class:`Vec2`. """
        return self._location

    @property
    def direction(self) -> Vec2:
        """ Direction vector as :class:`Vec2`. """
        return self._direction

    @property
    def slope(self) -> float:
        """ Slope of ray or ``None`` if vertical. """
        return self._slope

    @property
    def angle(self) -> float:
        """ Angle between x-axis and ray in radians. """
        return self._angle

    @property
    def angle_deg(self) -> float:
        """ Angle between x-axis and ray in degrees. """
        return math.degrees(self._angle)

    @property
    def is_vertical(self) -> bool:
        """ ``True`` if ray is vertical (parallel to y-axis). """
        return self._is_vertical

    @property
    def is_horizontal(self) -> bool:
        """ ``True`` if ray is horizontal (parallel to x-axis). """
        return self._is_horizontal

    def __repr__(self) -> str:
        return 'ConstructionRay(p1=({0.location.x:.3f}, {0.location.y:.3f}), ' \
               'angle={0.angle:.5f})'.format(self)

    def is_parallel(self, other: 'ConstructionRay') -> bool:
        """ Returns ``True`` if rays are parallel. """
        if self._is_vertical:
            return other._is_vertical
        if other._is_vertical:
            return False
        if self._is_horizontal:
            return other._is_horizontal
        return math.isclose(self._slope, other._slope, abs_tol=1e-12)

    def intersect(self, other: 'ConstructionRay') -> Vec2:
        """
        Returns the intersection point as ``(x, y)`` tuple of `self` and `other`.

        Raises:
             ParallelRaysError: if rays are parallel

        """
        ray1 = self
        ray2 = other
        if ray1.is_parallel(ray2):
            raise ParallelRaysError("Rays are parallel")

        if ray1._is_vertical:
            x = ray1._location.x
            if ray2.is_horizontal:
                y = ray2._location.y
            else:
                y = ray2.yof(x)
        elif ray2._is_vertical:
            x = ray2._location.x
            if ray1.is_horizontal:
                y = ray1._location.y
            else:
                y = ray1.yof(x)
        elif ray1._is_horizontal:
            y = ray1._location.y
            x = ray2.xof(y)
        elif ray2._is_horizontal:
            y = ray2._location.y
            x = ray1.xof(y)
        else:
            # calc intersection with the 'straight-line-equation'
            # based on y(x) = y0 + x*slope
            x = (ray1._yof0 - ray2._yof0) / (ray2._slope - ray1._slope)
            y = ray1.yof(x)
        return Vec2((x, y))

    def orthogonal(self, location: 'Vertex') -> 'ConstructionRay':
        """ Returns orthogonal ray at `location`. """
        return ConstructionRay(location, angle=self._angle + HALF_PI)

    def yof(self, x: float) -> float:
        """ Returns y-value of ray for `x` location.

        Raises:
            ArithmeticError: for vertical rays

        """
        if self._is_vertical:
            raise ArithmeticError
        return self._yof0 + float(x) * self._slope

    def xof(self, y: float) -> float:
        """ Returns x-value of ray for `y` location.

        Raises:
            ArithmeticError: for horizontal rays

        """
        if self._is_vertical:
            return self._location.x
        elif not self._is_horizontal:
            return (float(y) - self._yof0) / self._slope
        else:
            raise ArithmeticError

    def bisectrix(self, other: 'ConstructionRay') -> 'ConstructionRay':
        """ Bisectrix between `self` and `other`. """
        intersection = self.intersect(other)
        alpha = (self._angle + other._angle) / 2.
        return ConstructionRay(intersection, angle=alpha)


class ConstructionLine:
    """
    2D ConstructionLine is similar to :class:`ConstructionRay`, but has a start- and endpoint.
    The direction of line goes from start- to endpoint, "left of line" is always in relation
    to this line direction.

    Args:
        start: start point of line as :class:`Vec2` compatible object
        end: end point of line as :class:`Vec2` compatible object

    """

    def __init__(self, start: 'Vertex', end: 'Vertex'):
        self.start = Vec2(start)
        self.end = Vec2(end)

    def __repr__(self) -> str:
        return 'ConstructionLine({0.start}, {0.end})'.format(self)

    @property
    def bounding_box(self) -> BoundingBox2d:
        """ bounding box of line as :class:`BoundingBox2d` object. """
        return BoundingBox2d((self.start, self.end))

    def translate(self, dx: float, dy: float) -> None:
        """
        Move line about `dx` in x-axis and about `dy` in y-axis.

        Args:
            dx: translation in x-axis
            dy: translation in y-axis

        """
        v = Vec2(dx, dy)
        self.start += v
        self.end += v

    @property
    def sorted_points(self):
        return (self.end, self.start) if self.start > self.end else (
            self.start, self.end)

    @property
    def ray(self):
        """ collinear :class:`ConstructionRay`. """
        return ConstructionRay(self.start, self.end)

    def __eq__(self, other: 'ConstructionLine') -> bool:
        return self.sorted_points == other.sorted_points

    def __lt__(self, other: 'ConstructionLine') -> bool:
        return self.sorted_points < other.sorted_points

    def length(self) -> float:
        """ Returns length of line. """
        return (self.end - self.start).magnitude

    def midpoint(self) -> 'Vec2':
        """ Returns mid point of line. """
        return self.start.lerp(self.end)

    @property
    def is_vertical(self) -> bool:
        """ ``True`` if line is vertical. """
        return math.isclose(self.start.x, self.end.x)

    def inside_bounding_box(self, point: 'Vertex') -> bool:
        """ Returns ``True`` if `point` is inside of line bounding box. """
        return self.bounding_box.inside(point)

    def intersect(self, other: 'ConstructionLine',
                  abs_tol: float = TOLERANCE) -> Optional['Vec2']:
        """
        Returns the intersection point of to lines or ``None`` if they have no intersection point.

        Args:
            other: other :class:`ConstructionLine`
            abs_tol: tolerance for distance check

        """
        return intersection_line_line_2d(
            (self.start, self.end), (other.start, other.end), virtual=False,
            abs_tol=abs_tol)

    def has_intersection(self, other: 'ConstructionLine',
                         abs_tol: float = TOLERANCE) -> bool:
        """ Returns ``True`` if has intersection with `other` line. """
        return self.intersect(other, abs_tol=abs_tol) is not None

    def is_point_left_of_line(self, point: 'Vertex', colinear=False) -> bool:
        """
        Returns ``True`` if `point` is left of construction line in relation to the line direction from start to end.

        If `colinear` is ``True``, a colinear point is also left of the line.

        """
        return is_point_left_of_line(point, self.start, self.end,
                                     colinear=colinear)



# Created: 13.03.2011
# Copyright (c) 2011-2020, Manfred Moitzi
# License: MIT License
from typing import TYPE_CHECKING, Iterable, List, Iterator
from itertools import chain

from ezdxf.lldxf.tags import DXFStructureError
from ezdxf.entities import entity_linker

if TYPE_CHECKING:
    from ezdxf.eztypes import (
        TagWriter, Drawing, DXFEntity, Tags, DXFTagStorage, DXFGraphic,
        BlockRecord,
    )


class StoredSection:
    def __init__(self, entities: List['Tags']):
        self.entities = entities

    def export_dxf(self, tagwriter: 'TagWriter'):
        # (0, SECTION) (2, NAME) is stored in entities
        for entity in self.entities:
            tagwriter.write_tags(entity)
        # ENDSEC not stored in entities !!!
        tagwriter.write_str('  0\nENDSEC\n')


class EntitySection:
    """ :class:`EntitiesSection` is just a proxy for :class:`Modelspace` and
    active :class:`Paperspace` linked together.
    """

    def __init__(self, doc: 'Drawing' = None,
                 entities: Iterable['DXFEntity'] = None):
        self.doc = doc
        if entities is not None:
            self._build(iter(entities))

    def __iter__(self) -> Iterable['DXFEntity']:
        """ Iterable for all entities of modelspace and active paperspace. """
        layouts = self.doc.layouts
        for entity in chain(layouts.modelspace(), layouts.active_layout()):
            yield entity

    def __len__(self) -> int:
        """ Returns count of all entities of modelspace and active paperspace.
        """
        layouts = self.doc.layouts
        return len(layouts.modelspace()) + len(layouts.active_layout())

    # none public interface

    def _build(self, entities: Iterator['DXFEntity']) -> None:
        section_head: 'DXFTagStorage' = next(entities)
        if section_head.dxftype() != 'SECTION' or section_head.base_class[
            1] != (2, 'ENTITIES'):
            raise DXFStructureError(
                "Critical structure error in ENTITIES section.")

        def add(entity: 'DXFGraphic'):
            handle = entity.dxf.owner
            # higher priority for owner handle
            if handle == msp_layout_key:
                paperspace = 0
            elif handle == psp_layout_key:
                paperspace = 1
            else:  # paperspace flag as fallback
                paperspace = entity.dxf.paperspace

            if paperspace:
                psp.add_entity(entity)
            else:
                msp.add_entity(entity)

        msp: 'BlockRecord' = self.doc.block_records.get('*Model_Space')
        psp: 'BlockRecord' = self.doc.block_records.get('*Paper_Space')
        msp_layout_key = msp.dxf.handle
        psp_layout_key = psp.dxf.handle
        linked_entities = entity_linker()
        # Don't store linked entities (VERTEX, ATTRIB, SEQEND) in entity space
        for entity in entities:
            if not linked_entities(entity):
                add(entity)

    def export_dxf(self, tagwriter: 'TagWriter') -> None:
        layouts = self.doc.layouts
        tagwriter.write_str("  0\nSECTION\n  2\nENTITIES\n")
        # Just write *Model_Space and the active *Paper_Space into the
        # ENTITIES section.
        layouts.modelspace().entity_space.export_dxf(tagwriter)
        layouts.active_layout().entity_space.export_dxf(tagwriter)
        tagwriter.write_tag2(0, "ENDSEC")



# Copyright (c) 2019-2020 Manfred Moitzi
# License: MIT License
from typing import TYPE_CHECKING
from ezdxf.lldxf import validator
from ezdxf.lldxf.attributes import (
    DXFAttr, DXFAttributes, DefSubclass, XType, RETURN_DEFAULT,
    group_code_mapping
)
from ezdxf.lldxf.const import DXF12, SUBCLASS_MARKER
from ezdxf.math import NULLVEC, Z_AXIS
from ezdxf.math.transformtools import OCSTransform
from .dxfentity import base_class, SubclassProcessor
from .dxfgfx import DXFGraphic, acdb_entity, elevation_to_z_axis
from .factory import register_entity

if TYPE_CHECKING:
    from ezdxf.eztypes import TagWriter, DXFNamespace, Matrix44

__all__ = ['Shape']

acdb_shape = DefSubclass('AcDbShape', {
    # Elevation is a legacy feature from R11 and prior, do not use this
    # attribute, store the entity elevation in the z-axis of the vertices.
    # ezdxf does not export the elevation attribute!
    'elevation': DXFAttr(38, default=0, optional=True),

    # Thickness could be negative:
    'thickness': DXFAttr(39, default=0, optional=True),

    # Insertion point (in WCS)
    'insert': DXFAttr(10, xtype=XType.point3d, default=NULLVEC),

    # Shape size:
    'size': DXFAttr(40, default=1),

    # Shape name:
    'name': DXFAttr(2, default=''),

    # Rotation angle in degrees:
    'rotation': DXFAttr(50, default=0, optional=True),

    # Relative X scale factor
    'xscale': DXFAttr(
        41, default=1, optional=True,
        validator=validator.is_not_zero,
        fixer=RETURN_DEFAULT,
    ),
    # Oblique angle in degrees:
    'oblique': DXFAttr(51, default=0, optional=True),

    'extrusion': DXFAttr(
        210, xtype=XType.point3d, default=Z_AXIS, optional=True,
        validator=validator.is_not_null_vector,
        fixer=RETURN_DEFAULT,
    ),
})
acdb_shape_group_codes = group_code_mapping(acdb_shape)


@register_entity
class Shape(DXFGraphic):
    """ DXF SHAPE entity """
    DXFTYPE = 'SHAPE'
    DXFATTRIBS = DXFAttributes(base_class, acdb_entity, acdb_shape)

    def load_dxf_attribs(
            self, processor: SubclassProcessor = None) -> 'DXFNamespace':
        dxf = super().load_dxf_attribs(processor)
        if processor:
            processor.fast_load_dxfattribs(
                dxf, acdb_shape_group_codes, subclass=2, recover=True)
            if processor.r12:
                # Transform elevation attribute from R11 to z-axis values:
                elevation_to_z_axis(dxf, ('insert', ))
        return dxf

    def export_entity(self, tagwriter: 'TagWriter') -> None:
        """ Export entity specific data as DXF tags. """
        super().export_entity(tagwriter)
        if tagwriter.dxfversion > DXF12:
            tagwriter.write_tag2(SUBCLASS_MARKER, acdb_shape.name)
        self.dxf.export_dxf_attribs(tagwriter, [
            'insert', 'size', 'name', 'thickness', 'rotation', 'xscale',
            'oblique', 'extrusion',
        ])

    def transform(self, m: 'Matrix44') -> 'Shape':
        """ Transform SHAPE entity by transformation matrix `m` inplace.

        .. versionadded:: 0.13

        """
        dxf = self.dxf
        dxf.insert = m.transform(dxf.insert)  # DXF Reference: WCS?
        ocs = OCSTransform(self.dxf.extrusion, m)

        dxf.rotation = ocs.transform_deg_angle(dxf.rotation)
        dxf.size = ocs.transform_length((0, dxf.size, 0))
        dxf.x_scale = ocs.transform_length(
            (dxf.x_scale, 0, 0), reflection=dxf.x_scale)
        if dxf.hasattr('thickness'):
            dxf.thickness = ocs.transform_length(
                (0, 0, dxf.thickness), reflection=dxf.thickness)

        dxf.extrusion = ocs.new_extrusion
        return self



# Copyright (c) 2020, Manfred Moitzi
# License: MIT License
from typing import TYPE_CHECKING, Optional, Iterable, Tuple, List, Dict, cast
import sys
import struct
import math
from enum import IntEnum
from itertools import repeat
from ezdxf.lldxf import const
from ezdxf.tools.binarydata import bytes_to_hexstr, ByteStream, BitStream
from ezdxf import colors
from ezdxf.math import (
    Vec3, Matrix44, Z_AXIS, ConstructionCircle,
    ConstructionArc,
)
from ezdxf.entities import factory
import logging

if TYPE_CHECKING:
    from ezdxf.eztypes import (
        Tags, TagWriter, Drawing, Polymesh, Polyface, Polyline, Hatch,
    )

logger = logging.getLogger('ezdxf')

CHUNK_SIZE = 127


def load_proxy_graphic(tags: 'Tags', length_code: int = 160,
                       data_code: int = 310) -> Optional[bytes]:
    binary_data = [tag.value for tag in
                   tags.pop_tags(codes=(length_code, data_code)) if
                   tag.code == data_code]
    return b''.join(binary_data) if len(binary_data) else None


def export_proxy_graphic(data: bytes, tagwriter: 'TagWriter',
                         length_code: int = 160, data_code: int = 310) -> None:
    # Do not export proxy graphic for DXF R12 files
    assert tagwriter.dxfversion > const.DXF12

    length = len(data)
    if length == 0:
        return

    tagwriter.write_tag2(length_code, length)
    index = 0
    while index < length:
        hex_str = bytes_to_hexstr(data[index:index + CHUNK_SIZE])
        tagwriter.write_tag2(data_code, hex_str)
        index += CHUNK_SIZE


class ProxyGraphicTypes(IntEnum):
    EXTENTS = 1
    CIRCLE = 2
    CIRCLE_3P = 3
    CIRCULAR_ARC = 4
    CIRCULAR_ARC_3P = 5
    POLYLINE = 6
    POLYGON = 7
    MESH = 8
    SHELL = 9
    TEXT = 10
    TEXT2 = 11
    XLINE = 12
    RAY = 13
    ATTRIBUTE_COLOR = 14
    UNUSED_15 = 15
    ATTRIBUTE_LAYER = 16
    UNUSED_17 = 17
    ATTRIBUTE_LINETYPE = 18
    ATTRIBUTE_MARKER = 19
    ATTRIBUTE_FILL = 20
    UNUSED_21 = 21
    ATTRIBUTE_TRUE_COLOR = 22
    ATTRIBUTE_LINEWEIGHT = 23
    ATTRIBUTE_LTSCALE = 24
    ATTRIBUTE_THICKNESS = 25
    ATTRIBUTE_PLOT_STYLE_NAME = 26
    PUSH_CLIP = 27
    POP_CLIP = 28
    PUSH_MATRIX = 29
    PUSH_MATRIX2 = 30
    POP_MATRIX = 31
    POLYLINE_WITH_NORMALS = 32
    LWPOLYLINE = 33
    ATTRIBUTE_MATERIAL = 34
    ATTRIBUTE_MAPPER = 35
    UNICODE_TEXT = 36
    UNKNOWN_37 = 37
    UNICODE_TEXT2 = 38


class ProxyGraphic:
    def __init__(self, data: bytes, doc: 'Drawing' = None):
        self._doc = doc
        self._factory = factory.new
        self._buffer: bytes = data
        self._index: int = 8
        self.dxfversion = doc.dxfversion if doc else 'AC1015'
        self.color: int = const.BYLAYER
        self.layer: str = '0'
        self.linetype: str = 'BYLAYER'
        self.marker_index: int = 0
        self.fill: bool = False
        self.true_color: Optional[int] = None
        self.lineweight: int = const.LINEWEIGHT_DEFAULT
        self.ltscale: float = 1.0
        self.thickness: float = 0.0
        # Layer list in storage order
        self.layers: List[str] = []
        # Linetypes list in storage order
        self.linetypes: List[str] = []
        # List of text styles, with font name as key
        self.textstyles = dict()
        self.required_fonts = set()
        self.matrices = []

        if self._doc:
            self.layers = list(layer.dxf.name for layer in self._doc.layers)
            self.linetypes = list(
                linetype.dxf.name for linetype in self._doc.linetypes)
            self.textstyles = {style.dxf.font: style.dxf.name for style in
                               self._doc.styles}

    def info(self) -> Iterable[Tuple[int, int, str]]:
        index = self._index
        buffer = self._buffer
        while index < len(buffer):
            size, type_ = struct.unpack_from('<2L', self._buffer, offset=index)
            try:
                name = ProxyGraphicTypes(type_).name
            except ValueError:
                name = f'UNKNOWN_TYPE_{type_}'
            yield index, size, name
            index += size

    def virtual_entities(self):
        def transform(entity):
            if self.matrices:
                return entity.transform(self.matrices[-1])
            else:
                return entity

        index = self._index
        buffer = self._buffer
        while index < len(buffer):
            size, type_ = struct.unpack_from('<2L', self._buffer, offset=index)
            try:
                name = ProxyGraphicTypes(type_).name.lower()
            except ValueError:
                logger.debug(f'Unsupported Type Code: {type_}')
                index += size
                continue
            method = getattr(self, name, None)
            if method:
                result = method(self._buffer[index + 8: index + size])
                if isinstance(result, tuple):
                    for entity in result:
                        yield transform(entity)
                elif result:
                    yield transform(result)
                if result:  # reset fill after each graphic entity
                    self.fill = False
            else:
                logger.debug(f'Unsupported feature ProxyGraphic.{name}()')
            index += size

    def push_matrix(self, data: bytes):
        values = struct.unpack('<16d', data)
        m = Matrix44(values)
        m.transpose()
        self.matrices.append(m)

    def pop_matrix(self, data: bytes):
        if self.matrices:
            self.matrices.pop()

    def reset_colors(self):
        self.color = const.BYLAYER
        self.true_color = None

    def attribute_color(self, data: bytes):
        self.reset_colors()
        self.color = struct.unpack('<L', data)[0]
        if self.color < 0 or self.color > 256:
            self.color = const.BYLAYER

    def attribute_layer(self, data: bytes):
        if self._doc:
            index = struct.unpack('<L', data)[0]
            if index < len(self.layers):
                self.layer = self.layers[index]

    def attribute_linetype(self, data: bytes):
        if self._doc:
            index = struct.unpack('<L', data)[0]
            if index < len(self.linetypes):
                self.linetype = self.linetypes[index]

    def attribute_marker(self, data: bytes):
        self.marker_index = struct.unpack('<L', data)[0]

    def attribute_fill(self, data: bytes):
        self.fill = bool(struct.unpack('<L', data)[0])

    def attribute_true_color(self, data: bytes):
        self.reset_colors()
        code, value = colors.decode_raw_color(struct.unpack('<L', data)[0])
        if code == colors.COLOR_TYPE_RGB:
            self.true_color = colors.rgb2int(value)
        else:  # ACI colors, BYLAYER, BYBLOCK
            self.color = value

    def attribute_lineweight(self, data: bytes):
        lw = struct.unpack('<L', data)[0]
        if lw > const.MAX_VALID_LINEWEIGHT:
            self.lineweight = max(lw - 0x100000000, const.LINEWEIGHT_DEFAULT)
        else:
            self.lineweight = lw

    def attribute_ltscale(self, data: bytes):
        self.ltscale = struct.unpack('<d', data)[0]

    def attribute_thickness(self, data: bytes):
        self.thickness = struct.unpack('<d', data)[0]

    def circle(self, data: bytes):
        bs = ByteStream(data)
        attribs = self._build_dxf_attribs()
        attribs['center'] = Vec3(bs.read_vertex())
        attribs['radius'] = bs.read_float()
        attribs['extrusion'] = bs.read_vertex()
        return self._factory('CIRCLE', dxfattribs=attribs)

    def circle_3p(self, data: bytes):
        bs = ByteStream(data)
        attribs = self._build_dxf_attribs()
        p1 = Vec3(bs.read_vertex())
        p2 = Vec3(bs.read_vertex())
        p3 = Vec3(bs.read_vertex())
        circle = ConstructionCircle.from_3p(p1, p2, p3)
        attribs['center'] = circle.center
        attribs['radius'] = circle.radius
        return self._factory('CIRCLE', dxfattribs=attribs)

    def circular_arc(self, data: bytes):
        bs = ByteStream(data)
        attribs = self._build_dxf_attribs()
        attribs['center'] = Vec3(bs.read_vertex())
        attribs['radius'] = bs.read_float()
        normal = Vec3(bs.read_vertex())
        if normal != (0, 0, 1):
            logger.debug('ProxyGraphic: unsupported 3D ARC.')
        start_vec = Vec3(bs.read_vertex())
        sweep_angle = bs.read_float()
        arc_type = bs.read_struct('L')[0]
        # just do 2D for now
        start_angle = start_vec.angle_deg
        end_angle = start_angle + math.degrees(sweep_angle)
        attribs['start_angle'] = start_angle
        attribs['end_angle'] = end_angle
        return self._factory('ARC', dxfattribs=attribs)

    def circular_arc_3p(self, data: bytes):
        bs = ByteStream(data)
        attribs = self._build_dxf_attribs()
        p1 = Vec3(bs.read_vertex())
        p2 = Vec3(bs.read_vertex())
        p3 = Vec3(bs.read_vertex())
        arc_type = bs.read_struct('L')[0]
        arc = ConstructionArc.from_3p(p1, p3, p2)
        attribs['center'] = arc.center
        attribs['radius'] = arc.radius
        attribs['start_angle'] = arc.start_angle
        attribs['end_angle'] = arc.end_angle
        return self._factory('ARC', dxfattribs=attribs)

    def _filled_polygon(self, vertices, attribs):
        hatch = cast('Hatch', self._factory('HATCH', dxfattribs=attribs))
        hatch.paths.add_polyline_path(vertices, is_closed=True)
        return hatch

    def _polyline(self, vertices, *, close=False, normal=Z_AXIS):
        # Polyline without bulge values!
        # Current implementation ignores the normal vector!
        attribs = self._build_dxf_attribs()
        count = len(vertices)
        if count == 1 or (count == 2 and vertices[0].isclose(vertices[1])):
            attribs['location'] = vertices[0]
            return self._factory('POINT', dxfattribs=attribs)

        if self.fill and count > 2:
            polyline = self._filled_polygon(vertices, attribs)
        else:
            attribs['flags'] = const.POLYLINE_3D_POLYLINE
            polyline = cast('Polyline',
                            self._factory('POLYLINE', dxfattribs=attribs))
            polyline.append_vertices(vertices)
            if close:
                polyline.close()
        return polyline

    def polyline_with_normals(self, data: bytes):
        # Polyline without bulge values!
        vertices, normal = self._load_vertices(data, load_normal=True)
        return self._polyline(vertices, normal=normal)

    def polyline(self, data: bytes):
        # Polyline without bulge values!
        vertices, normal = self._load_vertices(data, load_normal=False)
        return self._polyline(vertices)

    def polygon(self, data: bytes):
        # Polyline without bulge values!
        vertices, normal = self._load_vertices(data, load_normal=False)
        return self._polyline(vertices, close=True)

    def lwpolyline(self, data: bytes):
        # OpenDesign Specs LWPLINE: 20.4.85 Page 211
        logger.warning(
            'Untested proxy graphic entity: LWPOLYLINE - Need examples!')
        bs = BitStream(data)
        flag = bs.read_bit_short()
        attribs = self._build_dxf_attribs()
        if flag & 4:
            attribs['const_width'] = bs.read_bit_double()
        if flag & 8:
            attribs['elevation'] = bs.read_bit_double()
        if flag & 2:
            attribs['thickness'] = bs.read_bit_double()
        if flag & 1:
            attribs['extrusion'] = Vec3(bs.read_bit_double(3))

        num_points = bs.read_bit_long()
        if flag & 16:
            num_bulges = bs.read_bit_long()
        else:
            num_bulges = 0

        if self.dxfversion >= 'AC1024':  # R2010+
            vertex_id_count = bs.read_bit_long()
        else:
            vertex_id_count = 0

        if flag & 32:
            num_width = bs.read_bit_long()
        else:
            num_width = 0
        # ignore DXF R13/14 special vertex order

        vertices = [bs.read_raw_double(2)]
        prev_point = vertices[-1]
        for _ in range(num_points - 1):
            x = bs.read_bit_double_default(default=prev_point[0])
            y = bs.read_bit_double_default(default=prev_point[1])
            prev_point = (x, y)
            vertices.append(prev_point)
        bulges = [bs.read_bit_double() for _ in range(num_bulges)]
        vertex_ids = [bs.read_bit_long() for _ in range(vertex_id_count)]
        widths = [(bs.read_bit_double(), bs.read_bit_double()) for _ in
                  range(num_width)]
        if len(bulges) == 0:
            bulges = list(repeat(0, num_points))
        if len(widths) == 0:
            widths = list(repeat((0, 0), num_points))
        points = []
        for v, w, b in zip(vertices, widths, bulges):
            points.append((v[0], v[1], w[0], w[1], b))
        lwpolyline = cast('LWPolyline',
                          self._factory('LWPOLYLINE', dxfattribs=attribs))
        lwpolyline.set_points(points)
        return lwpolyline

    def mesh(self, data: bytes):
        logger.warning('Untested proxy graphic entity: MESH - Need examples!')
        bs = ByteStream(data)
        rows, columns = bs.read_struct('2L')
        attribs = self._build_dxf_attribs()
        attribs['m_count'] = rows
        attribs['n_count'] = columns
        attribs['flags'] = const.POLYLINE_3D_POLYMESH
        polymesh = cast('Polymesh',
                        self._factory('POLYLINE', dxfattribs=attribs))
        polymesh.append_vertices(
            Vec3(bs.read_vertex()) for _ in range(rows * columns))
        return polymesh

    def shell(self, data: bytes):
        logger.warning('Untested proxy graphic entity: SHELL - Need examples!')
        bs = ByteStream(data)
        attribs = self._build_dxf_attribs()
        attribs['flags'] = const.POLYLINE_POLYFACE
        polyface = cast('Polyface',
                        self._factory('POLYLINE', dxfattribs=attribs))
        vertex_count = bs.read_long()
        vertices = [Vec3(bs.read_vertex()) for _ in range(vertex_count)]
        face_count = bs.read_long()
        faces = []
        for i in range(face_count):
            vertex_count = abs(bs.read_signed_long())
            face_indices = [bs.read_long() for _ in range(vertex_count)]
            face = [vertices[index] for index in face_indices]
            faces.append(face)
        polyface.append_faces(faces)
        polyface.optimize()
        # todo: SHELL - read face properties, but requires an example.
        return polyface

    def text(self, data: bytes):
        return self._text(data, unicode=False)

    def unicode_text(self, data: bytes):
        return self._text(data, unicode=True)

    def _text(self, data: bytes, unicode: bool = False):
        bs = ByteStream(data)
        start_point = Vec3(bs.read_vertex())
        normal = Vec3(bs.read_vertex())
        text_direction = Vec3(bs.read_vertex())
        height, width_factor, oblique_angle = bs.read_struct('<3d')
        if unicode:
            text = bs.read_padded_unicode_string()
        else:
            text = bs.read_padded_string()
        attribs = self._build_dxf_attribs()
        attribs['insert'] = start_point
        attribs['text'] = text
        attribs['height'] = height
        attribs['width'] = width_factor
        attribs['rotation'] = text_direction.angle_deg
        attribs['oblique'] = math.degrees(oblique_angle)
        attribs['extrusion'] = normal
        return self._factory('TEXT', dxfattribs=attribs)

    def text2(self, data: bytes):
        bs = ByteStream(data)
        start_point = Vec3(bs.read_vertex())
        normal = Vec3(bs.read_vertex())
        text_direction = Vec3(bs.read_vertex())
        text = bs.read_padded_string()
        ignore_length_of_string, raw = bs.read_struct('<2l')
        height, width_factor, oblique_angle, tracking_percentage = bs.read_struct(
            '<4d')
        is_backwards, is_upside_down, is_vertical, is_underline, is_overline = bs.read_struct(
            '<5L')
        font_filename = bs.read_padded_string()
        big_font_filename = bs.read_padded_string()
        attribs = self._build_dxf_attribs()
        attribs['insert'] = start_point
        attribs['text'] = text
        attribs['height'] = height
        attribs['width'] = width_factor
        attribs['rotation'] = text_direction.angle_deg
        attribs['oblique'] = math.degrees(oblique_angle)
        attribs['style'] = self._get_style(font_filename, big_font_filename)
        attribs['text_generation_flag'] = 2 * is_backwards + 4 * is_upside_down
        attribs['extrusion'] = normal
        return self._factory('TEXT', dxfattribs=attribs)

    def unicode_text2(self, data: bytes):
        bs = ByteStream(data)
        start_point = Vec3(bs.read_vertex())
        normal = Vec3(bs.read_vertex())
        text_direction = Vec3(bs.read_vertex())
        text = bs.read_padded_unicode_string()
        ignore_length_of_string, ignore_raw = bs.read_struct('<2l')
        height, width_factor, oblique_angle, tracking_percentage = bs.read_struct(
            '<4d')
        is_backwards, is_upside_down, is_vertical, is_underline, is_overline = bs.read_struct(
            '<5L')
        is_bold, is_italic, charset, pitch = bs.read_struct('<4L')
        type_face = bs.read_padded_unicode_string()
        font_filename = bs.read_padded_unicode_string()
        big_font_filename = bs.read_padded_unicode_string()
        attribs = self._build_dxf_attribs()
        attribs['insert'] = start_point
        attribs['text'] = text
        attribs['height'] = height
        attribs['width'] = width_factor
        attribs['rotation'] = text_direction.angle_deg
        attribs['oblique'] = math.degrees(oblique_angle)
        attribs['style'] = self._get_style(font_filename, big_font_filename)
        attribs['text_generation_flag'] = 2 * is_backwards + 4 * is_upside_down
        attribs['extrusion'] = normal
        return self._factory('TEXT', dxfattribs=attribs)

    def xline(self, data: bytes):
        return self._xline(data, 'XLINE')

    def ray(self, data: bytes):
        return self._xline(data, 'RAY')

    def _xline(self, data: bytes, type_: str):
        logger.warning(
            'Untested proxy graphic entity: RAY/XLINE - Need examples!')
        bs = ByteStream(data)
        attribs = self._build_dxf_attribs()
        start_point = Vec3(bs.read_vertex())
        other_point = Vec3(bs.read_vertex())
        attribs['start'] = start_point
        attribs['unit_vector'] = (other_point - start_point).normalize()
        return self._factory(type_, dxfattribs=attribs)

    def _get_style(self, font: str, bigfont: str) -> str:
        self.required_fonts.add(font)
        if font in self.textstyles:
            style = self.textstyles[font]
        else:
            style = font
            if self._doc:
                self._doc.styles.new(font, dxfattribs={'font': font,
                                                       'bigfont': bigfont})
        return style

    def _load_vertices(self, data: bytes, load_normal=False):
        normal = Z_AXIS
        bs = ByteStream(data)
        count = bs.read_struct('<L')[0]
        if load_normal:
            count += 1
        vertices = []
        while count > 0:
            vertices.append(Vec3(bs.read_struct('<3d')))
            count -= 1
        if load_normal:
            normal = vertices.pop()
        return vertices, normal

    def _build_dxf_attribs(self) -> Dict:
        attribs = dict()
        if self.layer != '0':
            attribs['layer'] = self.layer
        if self.color != const.BYLAYER:
            attribs['color'] = self.color
        if self.linetype != 'BYLAYER':
            attribs['linetype'] = self.linetype
        if self.lineweight != const.LINEWEIGHT_DEFAULT:
            attribs['lineweight'] = self.lineweight
        if self.ltscale != 1.0:
            attribs['ltscale'] = self.ltscale
        if self.true_color is not None:
            attribs['true_color'] = self.true_color
        return attribs


class ProxyGraphicDebugger(ProxyGraphic):
    def __init__(self, data: bytes, doc: 'Drawing' = None, debug_stream=None):
        super(ProxyGraphicDebugger, self).__init__(data, doc)
        if debug_stream is None:
            debug_stream = sys.stdout
        self._debug_stream = debug_stream

    def log_entities(self):
        self.log_separator(char='=', newline=False)
        self.log_message('Create virtual DXF entities:')
        self.log_separator(newline=False)
        for entity in self.virtual_entities():
            self.log_message(f"\n  * {entity.dxftype()}")
            self.log_message(f"  * {entity.graphic_properties()}\n")
        self.log_separator(char='=')

    def log_commands(self):
        self.log_separator(char='=', newline=False)
        self.log_message('Raw proxy commands:')
        self.log_separator(newline=False)
        for index, size, cmd in self.info():
            self.log_message(f"Command: {cmd} Index: {index} Size: {size}")
        self.log_separator(char='=')

    def log_separator(self, char='-', newline=True):
        self.log_message(char * 79)
        if newline:
            self.log_message('')

    def log_message(self, msg: str):
        print(msg, file=self._debug_stream)

    def log_state(self):
        self.log_message('> ' + self.get_state())

    def get_state(self) -> str:
        return f"ly: '{self.layer}', clr: {self.color}, lt: {self.linetype}, " \
               f"lw: {self.lineweight}, ltscale: {self.ltscale}, " \
               f"rgb: {self.true_color}, fill: {self.fill}"

    def attribute_color(self, data: bytes):
        self.log_message('Command: set COLOR')
        super().attribute_color(data)
        self.log_state()

    def attribute_layer(self, data: bytes):
        self.log_message('Command: set LAYER')
        super().attribute_layer(data)
        self.log_state()

    def attribute_linetype(self, data: bytes):
        self.log_message('Command: set LINETYPE')
        super().attribute_linetype(data)
        self.log_state()

    def attribute_true_color(self, data: bytes):
        self.log_message('Command: set TRUE-COLOR')
        super().attribute_true_color(data)
        self.log_state()

    def attribute_lineweight(self, data: bytes):
        self.log_message('Command: set LINEWEIGHT')
        super().attribute_lineweight(data)
        self.log_state()

    def attribute_ltscale(self, data: bytes):
        self.log_message('Command: set LTSCALE')
        super().attribute_ltscale(data)
        self.log_state()

    def attribute_fill(self, data: bytes):
        self.log_message('Command: set FILL')
        super().attribute_fill(data)
        self.log_state()



# Purpose: header variables factory
# Created: 20.11.2010
# Copyright (c) 2010-2018, Manfred Moitzi
# License: MIT License
from typing import Sequence, Union
from .types import DXFVertex, DXFTag, cast_tag_value


def SingleValue(value: Union[str, float], code: int = 1) -> DXFTag:
    return DXFTag(code, cast_tag_value(code, value))


def Point2D(value: Sequence[float]) -> DXFVertex:

    return DXFVertex(10, (value[0], value[1]))


def Point3D(value: Sequence[float]) -> DXFVertex:
    return DXFVertex(10, (value[0], value[1], value[2]))


class HeaderVarDef:
    def __init__(self, name, code, factory, mindxf, maxdxf, priority, default=None):
        self.name = name
        self.code = code
        self.factory = factory
        self.mindxf = mindxf
        self.maxdxf = maxdxf
        self.priority = priority
        self.default = default
